# This file is part of beets.
# Copyright 2016, Fabrice Laporte.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for the 'ftintitle' plugin."""

import unittest

from beets.test.helper import PluginTestCase
from beetsplug import ftintitle


class FtInTitlePluginFunctional(PluginTestCase):
    plugin = "ftintitle"

    def _ft_add_item(self, path, artist, title, aartist):
        return self.add_item(
            path=path,
            artist=artist,
            artist_sort=artist,
            title=title,
            albumartist=aartist,
        )

    def _ft_set_config(
        self, ftformat, drop=False, auto=True, keep_in_artist=False
    ):
        self.config["ftintitle"]["format"] = ftformat
        self.config["ftintitle"]["drop"] = drop
        self.config["ftintitle"]["auto"] = auto
        self.config["ftintitle"]["keep_in_artist"] = keep_in_artist

    def test_functional_drop(self):
        item = self._ft_add_item("/", "Alice ft Bob", "Song 1", "Alice")
        self.run_command("ftintitle", "-d")
        item.load()
        assert item["artist"] == "Alice"
        assert item["title"] == "Song 1"

    def test_functional_not_found(self):
        item = self._ft_add_item("/", "Alice ft Bob", "Song 1", "George")
        self.run_command("ftintitle", "-d")
        item.load()
        # item should be unchanged
        assert item["artist"] == "Alice ft Bob"
        assert item["title"] == "Song 1"

    def test_functional_custom_format(self):
        self._ft_set_config("feat. {0}")
        item = self._ft_add_item("/", "Alice ft Bob", "Song 1", "Alice")
        self.run_command("ftintitle")
        item.load()
        assert item["artist"] == "Alice"
        assert item["title"] == "Song 1 feat. Bob"

        self._ft_set_config("featuring {0}")
        item = self._ft_add_item("/", "Alice feat. Bob", "Song 1", "Alice")
        self.run_command("ftintitle")
        item.load()
        assert item["artist"] == "Alice"
        assert item["title"] == "Song 1 featuring Bob"

        self._ft_set_config("with {0}")
        item = self._ft_add_item("/", "Alice feat Bob", "Song 1", "Alice")
        self.run_command("ftintitle")
        item.load()
        assert item["artist"] == "Alice"
        assert item["title"] == "Song 1 with Bob"

    def test_functional_keep_in_artist(self):
        self._ft_set_config("feat. {0}", keep_in_artist=True)
        item = self._ft_add_item("/", "Alice ft Bob", "Song 1", "Alice")
        self.run_command("ftintitle")
        item.load()
        assert item["artist"] == "Alice ft Bob"
        assert item["title"] == "Song 1 feat. Bob"

        item = self._ft_add_item("/", "Alice ft Bob", "Song 1", "Alice")
        self.run_command("ftintitle", "-d")
        item.load()
        assert item["artist"] == "Alice ft Bob"
        assert item["title"] == "Song 1"


class FtInTitlePluginTest(unittest.TestCase):
    def setUp(self):
        """Set up configuration"""
        ftintitle.FtInTitlePlugin()

    def test_find_feat_part(self):
        test_cases = [
            {
                "artist": "Alice ft. Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice feat Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice featuring Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice & Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice and Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice With Bob",
                "album_artist": "Alice",
                "feat_part": "Bob",
            },
            {
                "artist": "Alice defeat Bob",
                "album_artist": "Alice",
                "feat_part": None,
            },
            {
                "artist": "Alice & Bob",
                "album_artist": "Bob",
                "feat_part": "Alice",
            },
            {
                "artist": "Alice ft. Bob",
                "album_artist": "Bob",
                "feat_part": "Alice",
            },
            {
                "artist": "Alice ft. Carol",
                "album_artist": "Bob",
                "feat_part": None,
            },
        ]

        for test_case in test_cases:
            feat_part = ftintitle.find_feat_part(
                test_case["artist"], test_case["album_artist"]
            )
            assert feat_part == test_case["feat_part"]

    def test_split_on_feat(self):
        parts = ftintitle.split_on_feat("Alice ft. Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice feat Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice feat. Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice featuring Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice & Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice and Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice With Bob")
        assert parts == ("Alice", "Bob")
        parts = ftintitle.split_on_feat("Alice defeat Bob")
        assert parts == ("Alice defeat Bob", None)

    def test_contains_feat(self):
        assert ftintitle.contains_feat("Alice ft. Bob")
        assert ftintitle.contains_feat("Alice feat. Bob")
        assert ftintitle.contains_feat("Alice feat Bob")
        assert ftintitle.contains_feat("Alice featuring Bob")
        assert ftintitle.contains_feat("Alice (ft. Bob)")
        assert ftintitle.contains_feat("Alice (feat. Bob)")
        assert ftintitle.contains_feat("Alice [ft. Bob]")
        assert ftintitle.contains_feat("Alice [feat. Bob]")
        assert not ftintitle.contains_feat("Alice defeat Bob")
        assert not ftintitle.contains_feat("Aliceft.Bob")
        assert not ftintitle.contains_feat("Alice (defeat Bob)")
        assert not ftintitle.contains_feat("Live and Let Go")
        assert not ftintitle.contains_feat("Come With Me")



# This file is part of beets.
# Copyright 2016, Thomas Scholtes.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.


import fnmatch
import os.path
import re
import sys
import unittest

import pytest
from mediafile import MediaFile

from beets import util
from beets.library import Item
from beets.test import _common
from beets.test.helper import (
    AsIsImporterMixin,
    ImportHelper,
    PluginTestCase,
    capture_log,
    control_stdin,
)
from beets.util import bytestring_path, displayable_path
from beetsplug import convert


def shell_quote(text):
    import shlex

    return shlex.quote(text)


class ConvertMixin:
    def tagged_copy_cmd(self, tag):
        """Return a conversion command that copies files and appends
        `tag` to the copy.
        """
        if re.search("[^a-zA-Z0-9]", tag):
            raise ValueError(
                "tag '{}' must only contain letters and digits".format(tag)
            )

        # A Python script that copies the file and appends a tag.
        stub = os.path.join(_common.RSRC, b"convert_stub.py").decode("utf-8")
        return "{} {} $source $dest {}".format(
            shell_quote(sys.executable), shell_quote(stub), tag
        )

    def assertFileTag(self, path, tag):
        """Assert that the path is a file and the files content ends
        with `tag`.
        """
        display_tag = tag
        tag = tag.encode("utf-8")
        self.assertIsFile(path)
        with open(path, "rb") as f:
            f.seek(-len(display_tag), os.SEEK_END)
            assert (
                f.read() == tag
            ), f"{displayable_path(path)} is not tagged with {display_tag}"

    def assertNoFileTag(self, path, tag):
        """Assert that the path is a file and the files content does not
        end with `tag`.
        """
        display_tag = tag
        tag = tag.encode("utf-8")
        self.assertIsFile(path)
        with open(path, "rb") as f:
            f.seek(-len(tag), os.SEEK_END)
            assert (
                f.read() != tag
            ), f"{displayable_path(path)} is unexpectedly tagged with {display_tag}"


class ConvertTestCase(ConvertMixin, PluginTestCase):
    db_on_disk = True
    plugin = "convert"


@_common.slow_test()
class ImportConvertTest(AsIsImporterMixin, ImportHelper, ConvertTestCase):
    def setUp(self):
        super().setUp()
        self.config["convert"] = {
            "dest": os.path.join(self.temp_dir, b"convert"),
            "command": self.tagged_copy_cmd("convert"),
            # Enforce running convert
            "max_bitrate": 1,
            "auto": True,
            "quiet": False,
        }

    def test_import_converted(self):
        self.run_asis_importer()
        item = self.lib.items().get()
        self.assertFileTag(item.path, "convert")

    # FIXME: fails on windows
    @unittest.skipIf(sys.platform == "win32", "win32")
    def test_import_original_on_convert_error(self):
        # `false` exits with non-zero code
        self.config["convert"]["command"] = "false"
        self.run_asis_importer()

        item = self.lib.items().get()
        assert item is not None
        self.assertIsFile(item.path)

    def test_delete_originals(self):
        self.config["convert"]["delete_originals"] = True
        self.run_asis_importer()
        for path in self.importer.paths:
            for root, dirnames, filenames in os.walk(path):
                assert (
                    len(fnmatch.filter(filenames, "*.mp3")) == 0
                ), f"Non-empty import directory {util.displayable_path(path)}"

    def get_count_of_import_files(self):
        import_file_count = 0

        for path in self.importer.paths:
            for root, _, filenames in os.walk(path):
                import_file_count += len(filenames)

        return import_file_count


class ConvertCommand:
    """A mixin providing a utility method to run the `convert`command
    in tests.
    """

    def run_convert_path(self, path, *args):
        """Run the `convert` command on a given path."""
        # The path is currently a filesystem bytestring. Convert it to
        # an argument bytestring.
        path = path.decode(util._fsencoding()).encode(util.arg_encoding())

        args = args + (b"path:" + path,)
        return self.run_command("convert", *args)

    def run_convert(self, *args):
        """Run the `convert` command on `self.item`."""
        return self.run_convert_path(self.item.path, *args)


@_common.slow_test()
class ConvertCliTest(ConvertTestCase, ConvertCommand):
    def setUp(self):
        super().setUp()
        self.album = self.add_album_fixture(ext="ogg")
        self.item = self.album.items()[0]

        self.convert_dest = bytestring_path(
            os.path.join(self.temp_dir, b"convert_dest")
        )
        self.config["convert"] = {
            "dest": self.convert_dest,
            "paths": {"default": "converted"},
            "format": "mp3",
            "formats": {
                "mp3": self.tagged_copy_cmd("mp3"),
                "ogg": self.tagged_copy_cmd("ogg"),
                "opus": {
                    "command": self.tagged_copy_cmd("opus"),
                    "extension": "ops",
                },
            },
        }

    def test_convert(self):
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_convert_with_auto_confirmation(self):
        self.run_convert("--yes")
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_reject_confirmation(self):
        with control_stdin("n"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertNotExists(converted)

    def test_convert_keep_new(self):
        assert os.path.splitext(self.item.path)[1] == b".ogg"

        with control_stdin("y"):
            self.run_convert("--keep-new")

        self.item.load()
        assert os.path.splitext(self.item.path)[1] == b".mp3"

    def test_format_option(self):
        with control_stdin("y"):
            self.run_convert("--format", "opus")
            converted = os.path.join(self.convert_dest, b"converted.ops")
        self.assertFileTag(converted, "opus")

    def test_embed_album_art(self):
        self.config["convert"]["embed"] = True
        image_path = os.path.join(_common.RSRC, b"image-2x3.jpg")
        self.album.artpath = image_path
        self.album.store()
        with open(os.path.join(image_path), "rb") as f:
            image_data = f.read()

        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        mediafile = MediaFile(converted)
        assert mediafile.images[0].data == image_data

    def test_skip_existing(self):
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.touch(converted, content="XXX")
        self.run_convert("--yes")
        with open(converted) as f:
            assert f.read() == "XXX"

    def test_pretend(self):
        self.run_convert("--pretend")
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertNotExists(converted)

    def test_empty_query(self):
        with capture_log("beets.convert") as logs:
            self.run_convert("An impossible query")
        assert logs[0] == "convert: Empty query result."

    def test_no_transcode_when_maxbr_set_high_and_different_formats(self):
        self.config["convert"]["max_bitrate"] = 5000
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_transcode_when_maxbr_set_low_and_different_formats(self):
        self.config["convert"]["max_bitrate"] = 5
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_transcode_when_maxbr_set_to_none_and_different_formats(self):
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_no_transcode_when_maxbr_set_high_and_same_formats(self):
        self.config["convert"]["max_bitrate"] = 5000
        self.config["convert"]["format"] = "ogg"
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.ogg")
        self.assertNoFileTag(converted, "ogg")

    def test_transcode_when_maxbr_set_low_and_same_formats(self):
        self.config["convert"]["max_bitrate"] = 5
        self.config["convert"]["format"] = "ogg"
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.ogg")
        self.assertFileTag(converted, "ogg")

    def test_transcode_when_maxbr_set_to_none_and_same_formats(self):
        self.config["convert"]["format"] = "ogg"
        with control_stdin("y"):
            self.run_convert()
        converted = os.path.join(self.convert_dest, b"converted.ogg")
        self.assertNoFileTag(converted, "ogg")

    def test_playlist(self):
        with control_stdin("y"):
            self.run_convert("--playlist", "playlist.m3u8")
            m3u_created = os.path.join(self.convert_dest, b"playlist.m3u8")
        assert os.path.exists(m3u_created)

    def test_playlist_pretend(self):
        self.run_convert("--playlist", "playlist.m3u8", "--pretend")
        m3u_created = os.path.join(self.convert_dest, b"playlist.m3u8")
        assert not os.path.exists(m3u_created)


@_common.slow_test()
class NeverConvertLossyFilesTest(ConvertTestCase, ConvertCommand):
    """Test the effect of the `never_convert_lossy_files` option."""

    def setUp(self):
        super().setUp()

        self.convert_dest = os.path.join(self.temp_dir, b"convert_dest")
        self.config["convert"] = {
            "dest": self.convert_dest,
            "paths": {"default": "converted"},
            "never_convert_lossy_files": True,
            "format": "mp3",
            "formats": {
                "mp3": self.tagged_copy_cmd("mp3"),
            },
        }

    def test_transcode_from_lossless(self):
        [item] = self.add_item_fixtures(ext="flac")
        with control_stdin("y"):
            self.run_convert_path(item.path)
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_transcode_from_lossy(self):
        self.config["convert"]["never_convert_lossy_files"] = False
        [item] = self.add_item_fixtures(ext="ogg")
        with control_stdin("y"):
            self.run_convert_path(item.path)
        converted = os.path.join(self.convert_dest, b"converted.mp3")
        self.assertFileTag(converted, "mp3")

    def test_transcode_from_lossy_prevented(self):
        [item] = self.add_item_fixtures(ext="ogg")
        with control_stdin("y"):
            self.run_convert_path(item.path)
        converted = os.path.join(self.convert_dest, b"converted.ogg")
        self.assertNoFileTag(converted, "mp3")


class TestNoConvert:
    """Test the effect of the `no_convert` option."""

    @pytest.mark.parametrize(
        "config_value, should_skip",
        [
            ("", False),
            ("bitrate:320", False),
            ("bitrate:320 format:ogg", False),
            ("bitrate:320 , format:ogg", True),
        ],
    )
    def test_no_convert_skip(self, config_value, should_skip):
        item = Item(format="ogg", bitrate=256)
        convert.config["convert"]["no_convert"] = config_value
        assert convert.in_no_convert(item) == should_skip



AUTHOR = "Adrian Sampson"

# General configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.extlinks"]

exclude_patterns = ["_build"]
source_suffix = ".rst"
master_doc = "index"

project = "beets"
copyright = "2016, Adrian Sampson"

version = "2.1"
release = "2.1.0"

pygments_style = "sphinx"

# External links to the bug tracker and other sites.
extlinks = {
    "bug": ("https://github.com/beetbox/beets/issues/%s", "#%s"),
    "user": ("https://github.com/%s", "%s"),
    "pypi": ("https://pypi.org/project/%s/", "%s"),
    "stdlib": ("https://docs.python.org/3/library/%s.html", "%s"),
}

linkcheck_ignore = [
    r"https://github.com/beetbox/beets/issues/",
    r"https://github.com/[^/]+$",  # ignore user pages
    r".*localhost.*",
    r"https?://127\.0\.0\.1",
    r"https://www.musixmatch.com/",  # blocks requests
    r"https://genius.com/",  # blocks requests
]

# Options for HTML output
htmlhelp_basename = "beetsdoc"

# Options for LaTeX output
latex_documents = [
    ("index", "beets.tex", "beets Documentation", AUTHOR, "manual"),
]

# Options for manual page output
man_pages = [
    (
        "reference/cli",
        "beet",
        "music tagger and library organizer",
        [AUTHOR],
        1,
    ),
    (
        "reference/config",
        "beetsconfig",
        "beets configuration file",
        [AUTHOR],
        5,
    ),
]

# Options for pydata theme
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "collapse_navigation": True,
    "logo": {
        "text": "beets",
    },
    "pygment_light_style": "bw",
}
html_title = "beets"
html_logo = "_static/beets_logo_nobg.png"
html_static_path = ["_static"]
html_css_files = ["beets.css"]



# This file is part of beets.
# Copyright 2016, Fabrice Laporte.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for the 'lastgenre' plugin."""

from unittest.mock import Mock

from beets import config
from beets.test import _common
from beets.test.helper import BeetsTestCase
from beetsplug import lastgenre


class LastGenrePluginTest(BeetsTestCase):
    def setUp(self):
        super().setUp()
        self.plugin = lastgenre.LastGenrePlugin()

    def _setup_config(
        self, whitelist=False, canonical=False, count=1, prefer_specific=False
    ):
        config["lastgenre"]["canonical"] = canonical
        config["lastgenre"]["count"] = count
        config["lastgenre"]["prefer_specific"] = prefer_specific
        if isinstance(whitelist, (bool, (str,))):
            # Filename, default, or disabled.
            config["lastgenre"]["whitelist"] = whitelist
        self.plugin.setup()
        if not isinstance(whitelist, (bool, (str,))):
            # Explicit list of genres.
            self.plugin.whitelist = whitelist

    def test_default(self):
        """Fetch genres with whitelist and c14n deactivated"""
        self._setup_config()
        assert self.plugin._resolve_genres(["delta blues"]) == "Delta Blues"

    def test_c14n_only(self):
        """Default c14n tree funnels up to most common genre except for *wrong*
        genres that stay unchanged.
        """
        self._setup_config(canonical=True, count=99)
        assert self.plugin._resolve_genres(["delta blues"]) == "Blues"
        assert self.plugin._resolve_genres(["iota blues"]) == "Iota Blues"

    def test_whitelist_only(self):
        """Default whitelist rejects *wrong* (non existing) genres."""
        self._setup_config(whitelist=True)
        assert self.plugin._resolve_genres(["iota blues"]) == ""

    def test_whitelist_c14n(self):
        """Default whitelist and c14n both activated result in all parents
        genres being selected (from specific to common).
        """
        self._setup_config(canonical=True, whitelist=True, count=99)
        assert (
            self.plugin._resolve_genres(["delta blues"]) == "Delta Blues, Blues"
        )

    def test_whitelist_custom(self):
        """Keep only genres that are in the whitelist."""
        self._setup_config(whitelist={"blues", "rock", "jazz"}, count=2)
        assert self.plugin._resolve_genres(["pop", "blues"]) == "Blues"

        self._setup_config(canonical="", whitelist={"rock"})
        assert self.plugin._resolve_genres(["delta blues"]) == ""

    def test_count(self):
        """Keep the n first genres, as we expect them to be sorted from more to
        less popular.
        """
        self._setup_config(whitelist={"blues", "rock", "jazz"}, count=2)
        assert (
            self.plugin._resolve_genres(["jazz", "pop", "rock", "blues"])
            == "Jazz, Rock"
        )

    def test_count_c14n(self):
        """Keep the n first genres, after having applied c14n when necessary"""
        self._setup_config(
            whitelist={"blues", "rock", "jazz"}, canonical=True, count=2
        )
        # thanks to c14n, 'blues' superseeds 'country blues' and takes the
        # second slot
        assert (
            self.plugin._resolve_genres(
                ["jazz", "pop", "country blues", "rock"]
            )
            == "Jazz, Blues"
        )

    def test_c14n_whitelist(self):
        """Genres first pass through c14n and are then filtered"""
        self._setup_config(canonical=True, whitelist={"rock"})
        assert self.plugin._resolve_genres(["delta blues"]) == ""

    def test_empty_string_enables_canonical(self):
        """For backwards compatibility, setting the `canonical` option
        to the empty string enables it using the default tree.
        """
        self._setup_config(canonical="", count=99)
        assert self.plugin._resolve_genres(["delta blues"]) == "Blues"

    def test_empty_string_enables_whitelist(self):
        """Again for backwards compatibility, setting the `whitelist`
        option to the empty string enables the default set of genres.
        """
        self._setup_config(whitelist="")
        assert self.plugin._resolve_genres(["iota blues"]) == ""

    def test_prefer_specific_loads_tree(self):
        """When prefer_specific is enabled but canonical is not the
        tree still has to be loaded.
        """
        self._setup_config(prefer_specific=True, canonical=False)
        assert self.plugin.c14n_branches != []

    def test_prefer_specific_without_canonical(self):
        """Prefer_specific works without canonical."""
        self._setup_config(prefer_specific=True, canonical=False, count=4)
        assert (
            self.plugin._resolve_genres(["math rock", "post-rock"])
            == "Post-Rock, Math Rock"
        )

    def test_no_duplicate(self):
        """Remove duplicated genres."""
        self._setup_config(count=99)
        assert self.plugin._resolve_genres(["blues", "blues"]) == "Blues"

    def test_tags_for(self):
        class MockPylastElem:
            def __init__(self, name):
                self.name = name

            def get_name(self):
                return self.name

        class MockPylastObj:
            def get_top_tags(self):
                tag1 = Mock()
                tag1.weight = 90
                tag1.item = MockPylastElem("Pop")
                tag2 = Mock()
                tag2.weight = 40
                tag2.item = MockPylastElem("Rap")
                return [tag1, tag2]

        plugin = lastgenre.LastGenrePlugin()
        res = plugin._tags_for(MockPylastObj())
        assert res == ["pop", "rap"]
        res = plugin._tags_for(MockPylastObj(), min_weight=50)
        assert res == ["pop"]

    def test_get_genre(self):
        mock_genres = {"track": "1", "album": "2", "artist": "3"}

        def mock_fetch_track_genre(self, obj=None):
            return mock_genres["track"]

        def mock_fetch_album_genre(self, obj):
            return mock_genres["album"]

        def mock_fetch_artist_genre(self, obj):
            return mock_genres["artist"]

        lastgenre.LastGenrePlugin.fetch_track_genre = mock_fetch_track_genre
        lastgenre.LastGenrePlugin.fetch_album_genre = mock_fetch_album_genre
        lastgenre.LastGenrePlugin.fetch_artist_genre = mock_fetch_artist_genre

        self._setup_config(whitelist=False)
        item = _common.item()
        item.genre = mock_genres["track"]

        config["lastgenre"] = {"force": False}
        res = self.plugin._get_genre(item)
        assert res == (item.genre, "keep")

        config["lastgenre"] = {"force": True, "source": "track"}
        res = self.plugin._get_genre(item)
        assert res == (mock_genres["track"], "track")

        config["lastgenre"] = {"source": "album"}
        res = self.plugin._get_genre(item)
        assert res == (mock_genres["album"], "album")

        config["lastgenre"] = {"source": "artist"}
        res = self.plugin._get_genre(item)
        assert res == (mock_genres["artist"], "artist")

        mock_genres["artist"] = None
        res = self.plugin._get_genre(item)
        assert res == (item.genre, "original")

        config["lastgenre"] = {"fallback": "rap"}
        item.genre = None
        res = self.plugin._get_genre(item)
        assert res == (config["lastgenre"]["fallback"].get(), "fallback")

    def test_sort_by_depth(self):
        self._setup_config(canonical=True)
        # Normal case.
        tags = ("electronic", "ambient", "post-rock", "downtempo")
        res = self.plugin._sort_by_depth(tags)
        assert res == ["post-rock", "downtempo", "ambient", "electronic"]
        # Non-canonical tag ('chillout') present.
        tags = ("electronic", "ambient", "chillout")
        res = self.plugin._sort_by_depth(tags)
        assert res == ["ambient", "electronic"]



import importlib.util
import os

import pytest

from beets.test.helper import ImportHelper, PluginMixin

github_ci = os.environ.get("GITHUB_ACTIONS") == "true"
if not github_ci and not importlib.util.find_spec("librosa"):
    pytest.skip("librosa isn't available", allow_module_level=True)


class TestAutoBPMPlugin(PluginMixin, ImportHelper):
    plugin = "autobpm"

    @pytest.fixture(scope="class", name="lib")
    def fixture_lib(self):
        self.setup_beets()

        yield self.lib

        self.teardown_beets()

    @pytest.fixture(scope="class")
    def item(self):
        return self.add_item_fixture()

    @pytest.fixture(scope="class")
    def importer(self, lib):
        self.import_media = []
        self.prepare_album_for_import(1)
        track = self.import_media[0]
        track.bpm = None
        track.save()
        return self.setup_importer(autotag=False)

    def test_command(self, lib, item):
        self.run_command("autobpm", lib=lib)

        item.load()
        assert item.bpm == 117

    def test_import(self, lib, importer):
        importer.run()

        assert lib.items().get().bpm == 117



# This file is part of beets.
# Copyright 2016, Adrian Sampson and Diego Moreda.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

import codecs
from unittest.mock import patch

from beets.dbcore.query import TrueQuery
from beets.library import Item
from beets.test import _common
from beets.test.helper import (
    AutotagStub,
    BeetsTestCase,
    ImportTestCase,
    PluginMixin,
    TerminalImportMixin,
    control_stdin,
)


class ModifyFileMocker:
    """Helper for modifying a file, replacing or editing its contents. Used for
    mocking the calls to the external editor during testing.
    """

    def __init__(self, contents=None, replacements=None):
        """`self.contents` and `self.replacements` are initialized here, in
        order to keep the rest of the functions of this class with the same
        signature as `EditPlugin.get_editor()`, making mocking easier.
            - `contents`: string with the contents of the file to be used for
            `overwrite_contents()`
            - `replacement`: dict with the in-place replacements to be used for
            `replace_contents()`, in the form {'previous string': 'new string'}

        TODO: check if it can be solved more elegantly with a decorator
        """
        self.contents = contents
        self.replacements = replacements
        self.action = self.overwrite_contents
        if replacements:
            self.action = self.replace_contents

    # The two methods below mock the `edit` utility function in the plugin.

    def overwrite_contents(self, filename, log):
        """Modify `filename`, replacing its contents with `self.contents`. If
        `self.contents` is empty, the file remains unchanged.
        """
        if self.contents:
            with codecs.open(filename, "w", encoding="utf-8") as f:
                f.write(self.contents)

    def replace_contents(self, filename, log):
        """Modify `filename`, reading its contents and replacing the strings
        specified in `self.replacements`.
        """
        with codecs.open(filename, "r", encoding="utf-8") as f:
            contents = f.read()
        for old, new_ in self.replacements.items():
            contents = contents.replace(old, new_)
        with codecs.open(filename, "w", encoding="utf-8") as f:
            f.write(contents)


class EditMixin(PluginMixin):
    """Helper containing some common functionality used for the Edit tests."""

    plugin = "edit"

    def assertItemFieldsModified(
        self, library_items, items, fields=[], allowed=["path"]
    ):
        """Assert that items in the library (`lib_items`) have different values
        on the specified `fields` (and *only* on those fields), compared to
        `items`.

        An empty `fields` list results in asserting that no modifications have
        been performed. `allowed` is a list of field changes that are ignored
        (they may or may not have changed; the assertion doesn't care).
        """
        for lib_item, item in zip(library_items, items):
            diff_fields = [
                field
                for field in lib_item._fields
                if lib_item[field] != item[field]
            ]
            assert set(diff_fields).difference(allowed) == set(fields)

    def run_mocked_interpreter(self, modify_file_args={}, stdin=[]):
        """Run the edit command during an import session, with mocked stdin and
        yaml writing.
        """
        m = ModifyFileMocker(**modify_file_args)
        with patch("beetsplug.edit.edit", side_effect=m.action):
            with control_stdin("\n".join(stdin)):
                self.importer.run()

    def run_mocked_command(self, modify_file_args={}, stdin=[], args=[]):
        """Run the edit command, with mocked stdin and yaml writing, and
        passing `args` to `run_command`."""
        m = ModifyFileMocker(**modify_file_args)
        with patch("beetsplug.edit.edit", side_effect=m.action):
            with control_stdin("\n".join(stdin)):
                self.run_command("edit", *args)


@_common.slow_test()
@patch("beets.library.Item.write")
class EditCommandTest(EditMixin, BeetsTestCase):
    """Black box tests for `beetsplug.edit`. Command line interaction is
    simulated using `test.helper.control_stdin()`, and yaml editing via an
    external editor is simulated using `ModifyFileMocker`.
    """

    ALBUM_COUNT = 1
    TRACK_COUNT = 10

    def setUp(self):
        super().setUp()
        # Add an album, storing the original fields for comparison.
        self.album = self.add_album_fixture(track_count=self.TRACK_COUNT)
        self.album_orig = {f: self.album[f] for f in self.album._fields}
        self.items_orig = [
            {f: item[f] for f in item._fields} for item in self.album.items()
        ]

    def assertCounts(
        self,
        mock_write,
        album_count=ALBUM_COUNT,
        track_count=TRACK_COUNT,
        write_call_count=TRACK_COUNT,
        title_starts_with="",
    ):
        """Several common assertions on Album, Track and call counts."""
        assert len(self.lib.albums()) == album_count
        assert len(self.lib.items()) == track_count
        assert mock_write.call_count == write_call_count
        assert all(
            i.title.startswith(title_starts_with) for i in self.lib.items()
        )

    def test_title_edit_discard(self, mock_write):
        """Edit title for all items in the library, then discard changes."""
        # Edit track titles.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # Cancel.
            ["c"],
        )

        self.assertCounts(
            mock_write, write_call_count=0, title_starts_with="t\u00eftle"
        )
        self.assertItemFieldsModified(self.album.items(), self.items_orig, [])

    def test_title_edit_apply(self, mock_write):
        """Edit title for all items in the library, then apply changes."""
        # Edit track titles.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # Apply changes.
            ["a"],
        )

        self.assertCounts(
            mock_write,
            write_call_count=self.TRACK_COUNT,
            title_starts_with="modified t\u00eftle",
        )
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["title", "mtime"]
        )

    def test_single_title_edit_apply(self, mock_write):
        """Edit title for one item in the library, then apply changes."""
        # Edit one track title.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle 9": "modified t\u00eftle 9"}},
            # Apply changes.
            ["a"],
        )

        self.assertCounts(
            mock_write,
            write_call_count=1,
        )
        # No changes except on last item.
        self.assertItemFieldsModified(
            list(self.album.items())[:-1], self.items_orig[:-1], []
        )
        assert list(self.album.items())[-1].title == "modified t\u00eftle 9"

    def test_noedit(self, mock_write):
        """Do not edit anything."""
        # Do not edit anything.
        self.run_mocked_command(
            {"contents": None},
            # No stdin.
            [],
        )

        self.assertCounts(
            mock_write, write_call_count=0, title_starts_with="t\u00eftle"
        )
        self.assertItemFieldsModified(self.album.items(), self.items_orig, [])

    def test_album_edit_apply(self, mock_write):
        """Edit the album field for all items in the library, apply changes.
        By design, the album should not be updated.""
        """
        # Edit album.
        self.run_mocked_command(
            {"replacements": {"\u00e4lbum": "modified \u00e4lbum"}},
            # Apply changes.
            ["a"],
        )

        self.assertCounts(mock_write, write_call_count=self.TRACK_COUNT)
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["album", "mtime"]
        )
        # Ensure album is *not* modified.
        self.album.load()
        assert self.album.album == "\u00e4lbum"

    def test_single_edit_add_field(self, mock_write):
        """Edit the yaml file appending an extra field to the first item, then
        apply changes."""
        # Append "foo: bar" to item with id == 2. ("id: 1" would match both
        # "id: 1" and "id: 10")
        self.run_mocked_command(
            {"replacements": {"id: 2": "id: 2\nfoo: bar"}},
            # Apply changes.
            ["a"],
        )

        assert self.lib.items("id:2")[0].foo == "bar"
        # Even though a flexible attribute was written (which is not directly
        # written to the tags), write should still be called since templates
        # might use it.
        self.assertCounts(
            mock_write, write_call_count=1, title_starts_with="t\u00eftle"
        )

    def test_a_album_edit_apply(self, mock_write):
        """Album query (-a), edit album field, apply changes."""
        self.run_mocked_command(
            {"replacements": {"\u00e4lbum": "modified \u00e4lbum"}},
            # Apply changes.
            ["a"],
            args=["-a"],
        )

        self.album.load()
        self.assertCounts(mock_write, write_call_count=self.TRACK_COUNT)
        assert self.album.album == "modified \u00e4lbum"
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["album", "mtime"]
        )

    def test_a_albumartist_edit_apply(self, mock_write):
        """Album query (-a), edit albumartist field, apply changes."""
        self.run_mocked_command(
            {"replacements": {"album artist": "modified album artist"}},
            # Apply changes.
            ["a"],
            args=["-a"],
        )

        self.album.load()
        self.assertCounts(mock_write, write_call_count=self.TRACK_COUNT)
        assert self.album.albumartist == "the modified album artist"
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["albumartist", "mtime"]
        )

    def test_malformed_yaml(self, mock_write):
        """Edit the yaml file incorrectly (resulting in a malformed yaml
        document)."""
        # Edit the yaml file to an invalid file.
        self.run_mocked_command(
            {"contents": "!MALFORMED"},
            # Edit again to fix? No.
            ["n"],
        )

        self.assertCounts(
            mock_write, write_call_count=0, title_starts_with="t\u00eftle"
        )

    def test_invalid_yaml(self, mock_write):
        """Edit the yaml file incorrectly (resulting in a well-formed but
        invalid yaml document)."""
        # Edit the yaml file to an invalid but parseable file.
        self.run_mocked_command(
            {"contents": "wellformed: yes, but invalid"},
            # No stdin.
            [],
        )

        self.assertCounts(
            mock_write, write_call_count=0, title_starts_with="t\u00eftle"
        )


@_common.slow_test()
class EditDuringImporterTestCase(
    EditMixin, TerminalImportMixin, ImportTestCase
):
    """TODO"""

    IGNORED = ["added", "album_id", "id", "mtime", "path"]

    def setUp(self):
        super().setUp()
        # Create some mediafiles, and store them for comparison.
        self.prepare_album_for_import(1)
        self.items_orig = [Item.from_path(f.path) for f in self.import_media]
        self.matcher = AutotagStub().install()
        self.matcher.matching = AutotagStub.GOOD

    def tearDown(self):
        super().tearDown()
        self.matcher.restore()


@_common.slow_test()
class EditDuringImporterNonSingletonTest(EditDuringImporterTestCase):
    def setUp(self):
        super().setUp()
        self.importer = self.setup_importer()

    def test_edit_apply_asis(self):
        """Edit the album field for all items in the library, apply changes,
        using the original item tags.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        # Check that only the 'title' field is modified.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            ["title"],
            self.IGNORED
            + [
                "albumartist",
                "mb_albumartistid",
                "mb_albumartistids",
            ],
        )
        assert all("Edited Track" in i.title for i in self.lib.items())

        # Ensure album is *not* fetched from a candidate.
        assert self.lib.albums()[0].mb_albumid == ""

    def test_edit_discard_asis(self):
        """Edit the album field for all items in the library, discard changes,
        using the original item tags.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Cancel, Use as-is.
            ["d", "c", "u"],
        )

        # Check that nothing is modified, the album is imported ASIS.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            [],
            self.IGNORED + ["albumartist", "mb_albumartistid"],
        )
        assert all("Tag Track" in i.title for i in self.lib.items())

        # Ensure album is *not* fetched from a candidate.
        assert self.lib.albums()[0].mb_albumid == ""

    def test_edit_apply_candidate(self):
        """Edit the album field for all items in the library, apply changes,
        using a candidate.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes.
            ["c", "1", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_retag_apply(self):
        """Import the album using a candidate, then retag and edit and apply
        changes.
        """
        self.run_mocked_interpreter(
            {},
            # 1, Apply changes.
            ["1", "a"],
        )

        # Retag and edit track titles.  On retag, the importer will reset items
        # ids but not the db connections.
        self.importer.paths = []
        self.importer.query = TrueQuery()
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_discard_candidate(self):
        """Edit the album field for all items in the library, discard changes,
        using a candidate.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes.
            ["c", "1", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_apply_candidate_singleton(self):
        """Edit the album field for all items in the library, apply changes,
        using a candidate and singleton mode.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes, aBort.
            ["c", "1", "a", "b"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())


@_common.slow_test()
class EditDuringImporterSingletonTest(EditDuringImporterTestCase):
    def setUp(self):
        super().setUp()
        self.importer = self.setup_singleton_importer()

    def test_edit_apply_asis_singleton(self):
        """Edit the album field for all items in the library, apply changes,
        using the original item tags and singleton mode.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Apply changes, aBort.
            ["d", "a", "b"],
        )

        # Check that only the 'title' field is modified.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            ["title"],
            self.IGNORED + ["albumartist", "mb_albumartistid"],
        )
        assert all("Edited Track" in i.title for i in self.lib.items())



# This file is part of beets.
# Copyright 2016, Bruno Cauet.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.


from os import fsdecode, path, remove
from shutil import rmtree
from tempfile import mkdtemp
from unittest.mock import MagicMock, Mock, PropertyMock

import pytest

from beets import config
from beets.dbcore import OrQuery
from beets.dbcore.query import FixedFieldSort, MultipleSort, NullSort
from beets.library import Album, Item, parse_query_string
from beets.test.helper import BeetsTestCase, PluginTestCase
from beets.ui import UserError
from beets.util import CHAR_REPLACE, bytestring_path, syspath
from beetsplug.smartplaylist import SmartPlaylistPlugin


class SmartPlaylistTest(BeetsTestCase):
    def test_build_queries(self):
        spl = SmartPlaylistPlugin()
        assert spl._matched_playlists is None
        assert spl._unmatched_playlists is None

        config["smartplaylist"]["playlists"].set([])
        spl.build_queries()
        assert spl._matched_playlists == set()
        assert spl._unmatched_playlists == set()

        config["smartplaylist"]["playlists"].set(
            [
                {"name": "foo", "query": "FOO foo"},
                {"name": "bar", "album_query": ["BAR bar1", "BAR bar2"]},
                {"name": "baz", "query": "BAZ baz", "album_query": "BAZ baz"},
            ]
        )
        spl.build_queries()
        assert spl._matched_playlists == set()
        foo_foo = parse_query_string("FOO foo", Item)
        baz_baz = parse_query_string("BAZ baz", Item)
        baz_baz2 = parse_query_string("BAZ baz", Album)
        bar_bar = OrQuery(
            (
                parse_query_string("BAR bar1", Album)[0],
                parse_query_string("BAR bar2", Album)[0],
            )
        )
        assert spl._unmatched_playlists == {
            ("foo", foo_foo, (None, None)),
            ("baz", baz_baz, baz_baz2),
            ("bar", (None, None), (bar_bar, None)),
        }

    def test_build_queries_with_sorts(self):
        spl = SmartPlaylistPlugin()
        config["smartplaylist"]["playlists"].set(
            [
                {"name": "no_sort", "query": "foo"},
                {"name": "one_sort", "query": "foo year+"},
                {"name": "only_empty_sorts", "query": ["foo", "bar"]},
                {"name": "one_non_empty_sort", "query": ["foo year+", "bar"]},
                {
                    "name": "multiple_sorts",
                    "query": ["foo year+", "bar genre-"],
                },
                {
                    "name": "mixed",
                    "query": ["foo year+", "bar", "baz genre+ id-"],
                },
            ]
        )

        spl.build_queries()
        sorts = {name: sort for name, (_, sort), _ in spl._unmatched_playlists}

        sort = FixedFieldSort  # short cut since we're only dealing with this
        assert sorts["no_sort"] == NullSort()
        assert sorts["one_sort"] == sort("year")
        assert sorts["only_empty_sorts"] is None
        assert sorts["one_non_empty_sort"] == sort("year")
        assert sorts["multiple_sorts"] == MultipleSort(
            [sort("year"), sort("genre", False)]
        )
        assert sorts["mixed"] == MultipleSort(
            [sort("year"), sort("genre"), sort("id", False)]
        )

    def test_matches(self):
        spl = SmartPlaylistPlugin()

        a = MagicMock(Album)
        i = MagicMock(Item)

        assert not spl.matches(i, None, None)
        assert not spl.matches(a, None, None)

        query = Mock()
        query.match.side_effect = {i: True}.__getitem__
        assert spl.matches(i, query, None)
        assert not spl.matches(a, query, None)

        a_query = Mock()
        a_query.match.side_effect = {a: True}.__getitem__
        assert not spl.matches(i, None, a_query)
        assert spl.matches(a, None, a_query)

        assert spl.matches(i, query, a_query)
        assert spl.matches(a, query, a_query)

    def test_db_changes(self):
        spl = SmartPlaylistPlugin()

        nones = None, None
        pl1 = "1", ("q1", None), nones
        pl2 = "2", ("q2", None), nones
        pl3 = "3", ("q3", None), nones

        spl._unmatched_playlists = {pl1, pl2, pl3}
        spl._matched_playlists = set()

        spl.matches = Mock(return_value=False)
        spl.db_change(None, "nothing")
        assert spl._unmatched_playlists == {pl1, pl2, pl3}
        assert spl._matched_playlists == set()

        spl.matches.side_effect = lambda _, q, __: q == "q3"
        spl.db_change(None, "matches 3")
        assert spl._unmatched_playlists == {pl1, pl2}
        assert spl._matched_playlists == {pl3}

        spl.matches.side_effect = lambda _, q, __: q == "q1"
        spl.db_change(None, "matches 3")
        assert spl._matched_playlists == {pl1, pl3}
        assert spl._unmatched_playlists == {pl2}

    def test_playlist_update(self):
        spl = SmartPlaylistPlugin()

        i = Mock(path=b"/tagada.mp3")
        i.evaluate_template.side_effect = lambda pl, _: pl.replace(
            b"$title", b"ta:ga:da"
        ).decode()

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = [pl]

        dir = bytestring_path(mkdtemp())
        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = fsdecode(dir)
        try:
            spl.update_playlists(lib)
        except Exception:
            rmtree(syspath(dir))
            raise

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = path.join(dir, b"ta_ga_da-my_playlist_.m3u")
        self.assertExists(m3u_filepath)
        with open(syspath(m3u_filepath), "rb") as f:
            content = f.read()
        rmtree(syspath(dir))

        assert content == b"/tagada.mp3\n"

    def test_playlist_update_output_extm3u(self):
        spl = SmartPlaylistPlugin()

        i = MagicMock()
        type(i).artist = PropertyMock(return_value="fake artist")
        type(i).title = PropertyMock(return_value="fake title")
        type(i).length = PropertyMock(return_value=300.123)
        type(i).path = PropertyMock(return_value=b"/tagada.mp3")
        i.evaluate_template.side_effect = lambda pl, _: pl.replace(
            b"$title",
            b"ta:ga:da",
        ).decode()

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = [pl]

        dir = bytestring_path(mkdtemp())
        config["smartplaylist"]["output"] = "extm3u"
        config["smartplaylist"]["prefix"] = "http://beets:8337/files"
        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = fsdecode(dir)
        try:
            spl.update_playlists(lib)
        except Exception:
            rmtree(syspath(dir))
            raise

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = path.join(dir, b"ta_ga_da-my_playlist_.m3u")
        self.assertExists(m3u_filepath)
        with open(syspath(m3u_filepath), "rb") as f:
            content = f.read()
        rmtree(syspath(dir))

        assert (
            content
            == b"#EXTM3U\n"
            + b"#EXTINF:300,fake artist - fake title\n"
            + b"http://beets:8337/files/tagada.mp3\n"
        )

    def test_playlist_update_output_extm3u_fields(self):
        spl = SmartPlaylistPlugin()

        i = MagicMock()
        type(i).artist = PropertyMock(return_value="Fake Artist")
        type(i).title = PropertyMock(return_value="fake Title")
        type(i).length = PropertyMock(return_value=300.123)
        type(i).path = PropertyMock(return_value=b"/tagada.mp3")
        a = {"id": 456, "genre": "Fake Genre"}
        i.__getitem__.side_effect = a.__getitem__
        i.evaluate_template.side_effect = lambda pl, _: pl.replace(
            b"$title",
            b"ta:ga:da",
        ).decode()

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = [pl]

        dir = bytestring_path(mkdtemp())
        config["smartplaylist"]["output"] = "extm3u"
        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = fsdecode(dir)
        config["smartplaylist"]["fields"] = ["id", "genre"]
        try:
            spl.update_playlists(lib)
        except Exception:
            rmtree(syspath(dir))
            raise

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = path.join(dir, b"ta_ga_da-my_playlist_.m3u")
        self.assertExists(m3u_filepath)
        with open(syspath(m3u_filepath), "rb") as f:
            content = f.read()
        rmtree(syspath(dir))

        assert (
            content
            == b"#EXTM3U\n"
            + b'#EXTINF:300 id="456" genre="Fake Genre",Fake Artist - fake Title\n'
            + b"/tagada.mp3\n"
        )

    def test_playlist_update_uri_format(self):
        spl = SmartPlaylistPlugin()

        i = MagicMock()
        type(i).id = PropertyMock(return_value=3)
        type(i).path = PropertyMock(return_value=b"/tagada.mp3")
        i.evaluate_template.side_effect = lambda pl, _: pl.replace(
            b"$title", b"ta:ga:da"
        ).decode()

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = [pl]

        dir = bytestring_path(mkdtemp())
        tpl = "http://beets:8337/item/$id/file"
        config["smartplaylist"]["uri_format"] = tpl
        config["smartplaylist"]["playlist_dir"] = fsdecode(dir)
        # The following options should be ignored when uri_format is set
        config["smartplaylist"]["relative_to"] = "/data"
        config["smartplaylist"]["prefix"] = "/prefix"
        config["smartplaylist"]["urlencode"] = True
        try:
            spl.update_playlists(lib)
        except Exception:
            rmtree(syspath(dir))
            raise

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = path.join(dir, b"ta_ga_da-my_playlist_.m3u")
        self.assertExists(m3u_filepath)
        with open(syspath(m3u_filepath), "rb") as f:
            content = f.read()
        rmtree(syspath(dir))

        assert content == b"http://beets:8337/item/3/file\n"


class SmartPlaylistCLITest(PluginTestCase):
    plugin = "smartplaylist"

    def setUp(self):
        super().setUp()

        self.item = self.add_item()
        config["smartplaylist"]["playlists"].set(
            [
                {"name": "my_playlist.m3u", "query": self.item.title},
                {"name": "all.m3u", "query": ""},
            ]
        )
        config["smartplaylist"]["playlist_dir"].set(fsdecode(self.temp_dir))

    def test_splupdate(self):
        with pytest.raises(UserError):
            self.run_with_output("splupdate", "tagada")

        self.run_with_output("splupdate", "my_playlist")
        m3u_path = path.join(self.temp_dir, b"my_playlist.m3u")
        self.assertExists(m3u_path)
        with open(syspath(m3u_path), "rb") as f:
            assert f.read() == self.item.path + b"\n"
        remove(syspath(m3u_path))

        self.run_with_output("splupdate", "my_playlist.m3u")
        with open(syspath(m3u_path), "rb") as f:
            assert f.read() == self.item.path + b"\n"
        remove(syspath(m3u_path))

        self.run_with_output("splupdate")
        for name in (b"my_playlist.m3u", b"all.m3u"):
            with open(path.join(self.temp_dir, name), "rb") as f:
                assert f.read() == self.item.path + b"\n"



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for discogs plugin."""

import pytest

from beets import config
from beets.test._common import Bag
from beets.test.helper import BeetsTestCase, capture_log
from beets.util.id_extractors import extract_discogs_id_regex
from beetsplug.discogs import DiscogsPlugin


class DGAlbumInfoTest(BeetsTestCase):
    def _make_release(self, tracks=None):
        """Returns a Bag that mimics a discogs_client.Release. The list
        of elements on the returned Bag is incomplete, including just
        those required for the tests on this class."""
        data = {
            "id": "ALBUM ID",
            "uri": "https://www.discogs.com/release/release/13633721",
            "title": "ALBUM TITLE",
            "year": "3001",
            "artists": [
                {"name": "ARTIST NAME", "id": "ARTIST ID", "join": ","}
            ],
            "formats": [
                {
                    "descriptions": ["FORMAT DESC 1", "FORMAT DESC 2"],
                    "name": "FORMAT",
                    "qty": 1,
                }
            ],
            "styles": ["STYLE1", "STYLE2"],
            "genres": ["GENRE1", "GENRE2"],
            "labels": [
                {
                    "name": "LABEL NAME",
                    "catno": "CATALOG NUMBER",
                }
            ],
            "tracklist": [],
        }

        if tracks:
            for recording in tracks:
                data["tracklist"].append(recording)

        return Bag(
            data=data,
            # Make some fields available as properties, as they are
            # accessed by DiscogsPlugin methods.
            title=data["title"],
            artists=[Bag(data=d) for d in data["artists"]],
        )

    def _make_track(self, title, position="", duration="", type_=None):
        track = {"title": title, "position": position, "duration": duration}
        if type_ is not None:
            # Test samples on discogs_client do not have a 'type_' field, but
            # the API seems to return it. Values: 'track' for regular tracks,
            # 'heading' for descriptive texts (ie. not real tracks - 12.13.2).
            track["type_"] = type_

        return track

    def _make_release_from_positions(self, positions):
        """Return a Bag that mimics a discogs_client.Release with a
        tracklist where tracks have the specified `positions`."""
        tracks = [
            self._make_track("TITLE%s" % i, position)
            for (i, position) in enumerate(positions, start=1)
        ]
        return self._make_release(tracks)

    def test_parse_media_for_tracks(self):
        tracks = [
            self._make_track("TITLE ONE", "1", "01:01"),
            self._make_track("TITLE TWO", "2", "02:02"),
        ]
        release = self._make_release(tracks=tracks)

        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks
        assert d.media == "FORMAT"
        assert t[0].media == d.media
        assert t[1].media == d.media

    def test_parse_medium_numbers_single_medium(self):
        release = self._make_release_from_positions(["1", "2"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 1
        assert t[0].medium == 1
        assert t[0].medium_total == 2
        assert t[1].medium == 1
        assert t[0].medium_total == 2

    def test_parse_medium_numbers_two_mediums(self):
        release = self._make_release_from_positions(["1-1", "2-1"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 2
        assert t[0].medium == 1
        assert t[0].medium_total == 1
        assert t[1].medium == 2
        assert t[1].medium_total == 1

    def test_parse_medium_numbers_two_mediums_two_sided(self):
        release = self._make_release_from_positions(["A1", "B1", "C1"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 2
        assert t[0].medium == 1
        assert t[0].medium_total == 2
        assert t[0].medium_index == 1
        assert t[1].medium == 1
        assert t[1].medium_total == 2
        assert t[1].medium_index == 2
        assert t[2].medium == 2
        assert t[2].medium_total == 1
        assert t[2].medium_index == 1

    def test_parse_track_indices(self):
        release = self._make_release_from_positions(["1", "2"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[0].medium_total == 2
        assert t[1].medium_index == 2
        assert t[1].index == 2
        assert t[1].medium_total == 2

    def test_parse_track_indices_several_media(self):
        release = self._make_release_from_positions(
            ["1-1", "1-2", "2-1", "3-1"]
        )
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 3
        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[0].medium_total == 2
        assert t[1].medium_index == 2
        assert t[1].index == 2
        assert t[1].medium_total == 2
        assert t[2].medium_index == 1
        assert t[2].index == 3
        assert t[2].medium_total == 1
        assert t[3].medium_index == 1
        assert t[3].index == 4
        assert t[3].medium_total == 1

    def test_parse_position(self):
        """Test the conversion of discogs `position` to medium, medium_index
        and subtrack_index."""
        # List of tuples (discogs_position, (medium, medium_index, subindex)
        positions = [
            ("1", (None, "1", None)),
            ("A12", ("A", "12", None)),
            ("12-34", ("12-", "34", None)),
            ("CD1-1", ("CD1-", "1", None)),
            ("1.12", (None, "1", "12")),
            ("12.a", (None, "12", "A")),
            ("12.34", (None, "12", "34")),
            ("1ab", (None, "1", "AB")),
            # Non-standard
            ("IV", ("IV", None, None)),
        ]

        d = DiscogsPlugin()
        for position, expected in positions:
            assert d.get_track_index(position) == expected

    def test_parse_tracklist_without_sides(self):
        """Test standard Discogs position 12.2.9#1: "without sides"."""
        release = self._make_release_from_positions(["1", "2", "3"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 3

    def test_parse_tracklist_with_sides(self):
        """Test standard Discogs position 12.2.9#2: "with sides"."""
        release = self._make_release_from_positions(["A1", "A2", "B1", "B2"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1  # 2 sides = 1 LP
        assert len(d.tracks) == 4

    def test_parse_tracklist_multiple_lp(self):
        """Test standard Discogs position 12.2.9#3: "multiple LP"."""
        release = self._make_release_from_positions(["A1", "A2", "B1", "C1"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 2  # 3 sides = 1 LP + 1 LP
        assert len(d.tracks) == 4

    def test_parse_tracklist_multiple_cd(self):
        """Test standard Discogs position 12.2.9#4: "multiple CDs"."""
        release = self._make_release_from_positions(
            ["1-1", "1-2", "2-1", "3-1"]
        )
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 3
        assert len(d.tracks) == 4

    def test_parse_tracklist_non_standard(self):
        """Test non standard Discogs position."""
        release = self._make_release_from_positions(["I", "II", "III", "IV"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 4

    def test_parse_tracklist_subtracks_dot(self):
        """Test standard Discogs position 12.2.9#5: "sub tracks, dots"."""
        release = self._make_release_from_positions(["1", "2.1", "2.2", "3"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 3

        release = self._make_release_from_positions(
            ["A1", "A2.1", "A2.2", "A3"]
        )
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 3

    def test_parse_tracklist_subtracks_letter(self):
        """Test standard Discogs position 12.2.9#5: "sub tracks, letter"."""
        release = self._make_release_from_positions(["A1", "A2a", "A2b", "A3"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 3

        release = self._make_release_from_positions(
            ["A1", "A2.a", "A2.b", "A3"]
        )
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 1
        assert len(d.tracks) == 3

    def test_parse_tracklist_subtracks_extra_material(self):
        """Test standard Discogs position 12.2.9#6: "extra material"."""
        release = self._make_release_from_positions(["1", "2", "Video 1"])
        d = DiscogsPlugin().get_album_info(release)

        assert d.mediums == 2
        assert len(d.tracks) == 3

    def test_parse_tracklist_subtracks_indices(self):
        """Test parsing of subtracks that include index tracks."""
        release = self._make_release_from_positions(["", "", "1.1", "1.2"])
        # Track 1: Index track with medium title
        release.data["tracklist"][0]["title"] = "MEDIUM TITLE"
        # Track 2: Index track with track group title
        release.data["tracklist"][1]["title"] = "TRACK GROUP TITLE"

        d = DiscogsPlugin().get_album_info(release)
        assert d.mediums == 1
        assert d.tracks[0].disctitle == "MEDIUM TITLE"
        assert len(d.tracks) == 1
        assert d.tracks[0].title == "TRACK GROUP TITLE"

    def test_parse_tracklist_subtracks_nested_logical(self):
        """Test parsing of subtracks defined inside a index track that are
        logical subtracks (ie. should be grouped together into a single track).
        """
        release = self._make_release_from_positions(["1", "", "3"])
        # Track 2: Index track with track group title, and sub_tracks
        release.data["tracklist"][1]["title"] = "TRACK GROUP TITLE"
        release.data["tracklist"][1]["sub_tracks"] = [
            self._make_track("TITLE ONE", "2.1", "01:01"),
            self._make_track("TITLE TWO", "2.2", "02:02"),
        ]

        d = DiscogsPlugin().get_album_info(release)
        assert d.mediums == 1
        assert len(d.tracks) == 3
        assert d.tracks[1].title == "TRACK GROUP TITLE"

    def test_parse_tracklist_subtracks_nested_physical(self):
        """Test parsing of subtracks defined inside a index track that are
        physical subtracks (ie. should not be grouped together).
        """
        release = self._make_release_from_positions(["1", "", "4"])
        # Track 2: Index track with track group title, and sub_tracks
        release.data["tracklist"][1]["title"] = "TRACK GROUP TITLE"
        release.data["tracklist"][1]["sub_tracks"] = [
            self._make_track("TITLE ONE", "2", "01:01"),
            self._make_track("TITLE TWO", "3", "02:02"),
        ]

        d = DiscogsPlugin().get_album_info(release)
        assert d.mediums == 1
        assert len(d.tracks) == 4
        assert d.tracks[1].title == "TITLE ONE"
        assert d.tracks[2].title == "TITLE TWO"

    def test_parse_tracklist_disctitles(self):
        """Test parsing of index tracks that act as disc titles."""
        release = self._make_release_from_positions(
            ["", "1-1", "1-2", "", "2-1"]
        )
        # Track 1: Index track with medium title (Cd1)
        release.data["tracklist"][0]["title"] = "MEDIUM TITLE CD1"
        # Track 4: Index track with medium title (Cd2)
        release.data["tracklist"][3]["title"] = "MEDIUM TITLE CD2"

        d = DiscogsPlugin().get_album_info(release)
        assert d.mediums == 2
        assert d.tracks[0].disctitle == "MEDIUM TITLE CD1"
        assert d.tracks[1].disctitle == "MEDIUM TITLE CD1"
        assert d.tracks[2].disctitle == "MEDIUM TITLE CD2"
        assert len(d.tracks) == 3

    def test_parse_minimal_release(self):
        """Test parsing of a release with the minimal amount of information."""
        data = {
            "id": 123,
            "uri": "https://www.discogs.com/release/123456-something",
            "tracklist": [self._make_track("A", "1", "01:01")],
            "artists": [{"name": "ARTIST NAME", "id": 321, "join": ""}],
            "title": "TITLE",
        }
        release = Bag(
            data=data,
            title=data["title"],
            artists=[Bag(data=d) for d in data["artists"]],
        )
        d = DiscogsPlugin().get_album_info(release)
        assert d.artist == "ARTIST NAME"
        assert d.album == "TITLE"
        assert len(d.tracks) == 1

    def test_parse_release_without_required_fields(self):
        """Test parsing of a release that does not have the required fields."""
        release = Bag(data={}, refresh=lambda *args: None)
        with capture_log() as logs:
            d = DiscogsPlugin().get_album_info(release)

        assert d is None
        assert "Release does not contain the required fields" in logs[0]

    def test_album_for_id(self):
        """Test parsing for a valid Discogs release_id"""
        test_patterns = [
            (
                "http://www.discogs.com/G%C3%BCnther-Lause-Meru-Ep/release/4354798",
                4354798,
            ),
            (
                "http://www.discogs.com/release/4354798-G%C3%BCnther-Lause-Meru-Ep",
                4354798,
            ),
            (
                "http://www.discogs.com/G%C3%BCnther-4354798Lause-Meru-Ep/release/4354798",  # NOQA E501
                4354798,
            ),
            (
                "http://www.discogs.com/release/4354798-G%C3%BCnther-4354798Lause-Meru-Ep/",  # NOQA E501
                4354798,
            ),
            ("[r4354798]", 4354798),
            ("r4354798", 4354798),
            ("4354798", 4354798),
            ("yet-another-metadata-provider.org/foo/12345", ""),
            ("005b84a0-ecd6-39f1-b2f6-6eb48756b268", ""),
        ]
        for test_pattern, expected in test_patterns:
            match = extract_discogs_id_regex(test_pattern)
            if not match:
                match = ""
            assert match == expected

    def test_default_genre_style_settings(self):
        """Test genre default settings, genres to genre, styles to style"""
        release = self._make_release_from_positions(["1", "2"])

        d = DiscogsPlugin().get_album_info(release)
        assert d.genre == "GENRE1, GENRE2"
        assert d.style == "STYLE1, STYLE2"

    def test_append_style_to_genre(self):
        """Test appending style to genre if config enabled"""
        config["discogs"]["append_style_genre"] = True
        release = self._make_release_from_positions(["1", "2"])

        d = DiscogsPlugin().get_album_info(release)
        assert d.genre == "GENRE1, GENRE2, STYLE1, STYLE2"
        assert d.style == "STYLE1, STYLE2"

    def test_append_style_to_genre_no_style(self):
        """Test nothing appended to genre if style is empty"""
        config["discogs"]["append_style_genre"] = True
        release = self._make_release_from_positions(["1", "2"])
        release.data["styles"] = []

        d = DiscogsPlugin().get_album_info(release)
        assert d.genre == "GENRE1, GENRE2"
        assert d.style is None


@pytest.mark.parametrize(
    "formats, expected_media, expected_albumtype",
    [
        (None, None, None),
        (
            [
                {
                    "descriptions": ['7"', "Single", "45 RPM"],
                    "name": "Vinyl",
                    "qty": 1,
                }
            ],
            "Vinyl",
            '7", Single, 45 RPM',
        ),
    ],
)
def test_get_media_and_albumtype(formats, expected_media, expected_albumtype):
    result = DiscogsPlugin.get_media_and_albumtype(formats)

    assert result == (expected_media, expected_albumtype)



# This file is part of beets.
# Copyright 2019, Carl Suster
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Test the beets.random utilities associated with the random plugin."""

import math
import unittest
from random import Random

import pytest

from beets import random
from beets.test.helper import TestHelper


class RandomTest(TestHelper, unittest.TestCase):
    def setUp(self):
        self.lib = None
        self.artist1 = "Artist 1"
        self.artist2 = "Artist 2"
        self.item1 = self.create_item(artist=self.artist1)
        self.item2 = self.create_item(artist=self.artist2)
        self.items = [self.item1, self.item2]
        for _ in range(8):
            self.items.append(self.create_item(artist=self.artist2))
        self.random_gen = Random()
        self.random_gen.seed(12345)

    def _stats(self, data):
        mean = sum(data) / len(data)
        stdev = math.sqrt(sum((p - mean) ** 2 for p in data) / (len(data) - 1))
        quot, rem = divmod(len(data), 2)
        if rem:
            median = sorted(data)[quot]
        else:
            median = sum(sorted(data)[quot - 1 : quot + 1]) / 2
        return mean, stdev, median

    def test_equal_permutation(self):
        """We have a list of items where only one item is from artist1 and the
        rest are from artist2. If we permute weighted by the artist field then
        the solo track will almost always end up near the start. If we use a
        different field then it'll be in the middle on average.
        """

        def experiment(field, histogram=False):
            """Permutes the list of items 500 times and calculates the position
            of self.item1 each time. Returns stats about that position.
            """
            positions = []
            for _ in range(500):
                shuffled = list(
                    random._equal_chance_permutation(
                        self.items, field=field, random_gen=self.random_gen
                    )
                )
                positions.append(shuffled.index(self.item1))
            # Print a histogram (useful for debugging).
            if histogram:
                for i in range(len(self.items)):
                    print("{:2d} {}".format(i, "*" * positions.count(i)))
            return self._stats(positions)

        mean1, stdev1, median1 = experiment("artist")
        mean2, stdev2, median2 = experiment("track")
        assert 0 == pytest.approx(median1, abs=1)
        assert len(self.items) // 2 == pytest.approx(median2, abs=1)
        assert stdev2 > stdev1



# This file is part of beets.
# Copyright 2020, David Swarbrick.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for image resizing based on filesize."""

import os
import unittest
from unittest.mock import patch

from beets.test import _common
from beets.test.helper import BeetsTestCase, CleanupModulesMixin
from beets.util import command_output, syspath
from beets.util.artresizer import IMBackend, PILBackend


class DummyIMBackend(IMBackend):
    """An `IMBackend` which pretends that ImageMagick is available.

    The version is sufficiently recent to support image comparison.
    """

    def __init__(self):
        """Init a dummy backend class for mocked ImageMagick tests."""
        self.version = (7, 0, 0)
        self.legacy = False
        self.convert_cmd = ["magick"]
        self.identify_cmd = ["magick", "identify"]
        self.compare_cmd = ["magick", "compare"]


class DummyPILBackend(PILBackend):
    """An `PILBackend` which pretends that PIL is available."""

    def __init__(self):
        """Init a dummy backend class for mocked PIL tests."""
        pass


class ArtResizerFileSizeTest(CleanupModulesMixin, BeetsTestCase):
    """Unittest test case for Art Resizer to a specific filesize."""

    modules = (IMBackend.__module__,)

    IMG_225x225 = os.path.join(_common.RSRC, b"abbey.jpg")
    IMG_225x225_SIZE = os.stat(syspath(IMG_225x225)).st_size

    def _test_img_resize(self, backend):
        """Test resizing based on file size, given a resize_func."""
        # Check quality setting unaffected by new parameter
        im_95_qual = backend.resize(
            225,
            self.IMG_225x225,
            quality=95,
            max_filesize=0,
        )
        # check valid path returned - max_filesize hasn't broken resize command
        self.assertExists(im_95_qual)

        # Attempt a lower filesize with same quality
        im_a = backend.resize(
            225,
            self.IMG_225x225,
            quality=95,
            max_filesize=0.9 * os.stat(syspath(im_95_qual)).st_size,
        )
        self.assertExists(im_a)
        # target size was achieved
        assert (
            os.stat(syspath(im_a)).st_size
            < os.stat(syspath(im_95_qual)).st_size
        )

        # Attempt with lower initial quality
        im_75_qual = backend.resize(
            225,
            self.IMG_225x225,
            quality=75,
            max_filesize=0,
        )
        self.assertExists(im_75_qual)

        im_b = backend.resize(
            225,
            self.IMG_225x225,
            quality=95,
            max_filesize=0.9 * os.stat(syspath(im_75_qual)).st_size,
        )
        self.assertExists(im_b)
        # Check high (initial) quality still gives a smaller filesize
        assert (
            os.stat(syspath(im_b)).st_size
            < os.stat(syspath(im_75_qual)).st_size
        )

    @unittest.skipUnless(PILBackend.available(), "PIL not available")
    def test_pil_file_resize(self):
        """Test PIL resize function is lowering file size."""
        self._test_img_resize(PILBackend())

    @unittest.skipUnless(IMBackend.available(), "ImageMagick not available")
    def test_im_file_resize(self):
        """Test IM resize function is lowering file size."""
        self._test_img_resize(IMBackend())

    @unittest.skipUnless(PILBackend.available(), "PIL not available")
    def test_pil_file_deinterlace(self):
        """Test PIL deinterlace function.

        Check if the `PILBackend.deinterlace()` function returns images
        that are non-progressive
        """
        path = PILBackend().deinterlace(self.IMG_225x225)
        from PIL import Image

        with Image.open(path) as img:
            assert "progression" not in img.info

    @unittest.skipUnless(IMBackend.available(), "ImageMagick not available")
    def test_im_file_deinterlace(self):
        """Test ImageMagick deinterlace function.

        Check if the `IMBackend.deinterlace()` function returns images
        that are non-progressive.
        """
        im = IMBackend()
        path = im.deinterlace(self.IMG_225x225)
        cmd = im.identify_cmd + [
            "-format",
            "%[interlace]",
            syspath(path, prefix=False),
        ]
        out = command_output(cmd).stdout
        assert out == b"None"

    @patch("beets.util.artresizer.util")
    def test_write_metadata_im(self, mock_util):
        """Test writing image metadata."""
        metadata = {"a": "A", "b": "B"}
        im = DummyIMBackend()
        im.write_metadata("foo", metadata)
        try:
            command = im.convert_cmd + "foo -set a A -set b B foo".split()
            mock_util.command_output.assert_called_once_with(command)
        except AssertionError:
            command = im.convert_cmd + "foo -set b B -set a A foo".split()
            mock_util.command_output.assert_called_once_with(command)



# This file is part of beets.
# Copyright 2019, Carl Suster
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Test the beets.export utilities associated with the export plugin."""

import json
import re  # used to test csv format
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from beets.test.helper import PluginTestCase


class ExportPluginTest(PluginTestCase):
    plugin = "export"

    def setUp(self):
        super().setUp()
        self.test_values = {"title": "xtitle", "album": "xalbum"}

    def execute_command(self, format_type, artist):
        query = ",".join(self.test_values.keys())
        out = self.run_with_output(
            "export", "-f", format_type, "-i", query, artist
        )
        return out

    def create_item(self):
        (item,) = self.add_item_fixtures()
        item.artist = "xartist"
        item.title = self.test_values["title"]
        item.album = self.test_values["album"]
        item.write()
        item.store()
        return item

    def test_json_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="json", artist=item1.artist)
        json_data = json.loads(out)[0]
        for key, val in self.test_values.items():
            assert key in json_data
            assert val == json_data[key]

    def test_jsonlines_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="jsonlines", artist=item1.artist)
        json_data = json.loads(out)
        for key, val in self.test_values.items():
            assert key in json_data
            assert val == json_data[key]

    def test_csv_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="csv", artist=item1.artist)
        csv_list = re.split("\r", re.sub("\n", "", out))
        head = re.split(",", csv_list[0])
        vals = re.split(",|\r", csv_list[1])
        for index, column in enumerate(head):
            assert self.test_values.get(column, None) is not None
            assert vals[index] == self.test_values[column]

    def test_xml_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="xml", artist=item1.artist)
        library = ElementTree.fromstring(out)
        assert isinstance(library, Element)
        for track in library[0]:
            for details in track:
                tag = details.tag
                txt = details.text
                assert tag in self.test_values, tag
                assert self.test_values[tag] == txt, txt



# This file is part of beets.
# Copyright 2016, Jesse Weinstein
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for the play plugin"""

import os
import sys
import unittest
from unittest.mock import ANY, patch

import pytest

from beets.test.helper import CleanupModulesMixin, PluginTestCase, control_stdin
from beets.ui import UserError
from beets.util import open_anything
from beetsplug.play import PlayPlugin


@patch("beetsplug.play.util.interactive_open")
class PlayPluginTest(CleanupModulesMixin, PluginTestCase):
    modules = (PlayPlugin.__module__,)
    plugin = "play"

    def setUp(self):
        super().setUp()
        self.item = self.add_item(album="a nice älbum", title="aNiceTitle")
        self.lib.add_album([self.item])
        self.config["play"]["command"] = "echo"

    def run_and_assert(
        self,
        open_mock,
        args=("title:aNiceTitle",),
        expected_cmd="echo",
        expected_playlist=None,
    ):
        self.run_command("play", *args)

        open_mock.assert_called_once_with(ANY, expected_cmd)
        expected_playlist = expected_playlist or self.item.path.decode("utf-8")
        exp_playlist = expected_playlist + "\n"
        with open(open_mock.call_args[0][0][0], "rb") as playlist:
            assert exp_playlist == playlist.read().decode("utf-8")

    def test_basic(self, open_mock):
        self.run_and_assert(open_mock)

    def test_album_option(self, open_mock):
        self.run_and_assert(open_mock, ["-a", "nice"])

    def test_args_option(self, open_mock):
        self.run_and_assert(
            open_mock, ["-A", "foo", "title:aNiceTitle"], "echo foo"
        )

    def test_args_option_in_middle(self, open_mock):
        self.config["play"]["command"] = "echo $args other"

        self.run_and_assert(
            open_mock, ["-A", "foo", "title:aNiceTitle"], "echo foo other"
        )

    def test_unset_args_option_in_middle(self, open_mock):
        self.config["play"]["command"] = "echo $args other"

        self.run_and_assert(open_mock, ["title:aNiceTitle"], "echo other")

    # FIXME: fails on windows
    @unittest.skipIf(sys.platform == "win32", "win32")
    def test_relative_to(self, open_mock):
        self.config["play"]["command"] = "echo"
        self.config["play"]["relative_to"] = "/something"

        path = os.path.relpath(self.item.path, b"/something")
        playlist = path.decode("utf-8")
        self.run_and_assert(
            open_mock, expected_cmd="echo", expected_playlist=playlist
        )

    def test_use_folders(self, open_mock):
        self.config["play"]["command"] = None
        self.config["play"]["use_folders"] = True
        self.run_command("play", "-a", "nice")

        open_mock.assert_called_once_with(ANY, open_anything())
        with open(open_mock.call_args[0][0][0], "rb") as f:
            playlist = f.read().decode("utf-8")
        assert (
            f'{os.path.dirname(self.item.path.decode("utf-8"))}\n' == playlist
        )

    def test_raw(self, open_mock):
        self.config["play"]["raw"] = True

        self.run_command("play", "nice")

        open_mock.assert_called_once_with([self.item.path], "echo")

    def test_not_found(self, open_mock):
        self.run_command("play", "not found")

        open_mock.assert_not_called()

    def test_warning_threshold(self, open_mock):
        self.config["play"]["warning_threshold"] = 1
        self.add_item(title="another NiceTitle")

        with control_stdin("a"):
            self.run_command("play", "nice")

        open_mock.assert_not_called()

    def test_skip_warning_threshold_bypass(self, open_mock):
        self.config["play"]["warning_threshold"] = 1
        self.other_item = self.add_item(title="another NiceTitle")

        expected_playlist = "{}\n{}".format(
            self.item.path.decode("utf-8"), self.other_item.path.decode("utf-8")
        )

        with control_stdin("a"):
            self.run_and_assert(
                open_mock,
                ["-y", "NiceTitle"],
                expected_playlist=expected_playlist,
            )

    def test_command_failed(self, open_mock):
        open_mock.side_effect = OSError("some reason")

        with pytest.raises(UserError):
            self.run_command("play", "title:aNiceTitle")



#!/usr/bin/env python3

"""A utility script for automating the beets release process."""

from __future__ import annotations

import re
import subprocess
from datetime import datetime, timezone
from pathlib import Path
from typing import Callable

import click
import tomli
from packaging.version import Version, parse

BASE = Path(__file__).parent.parent.absolute()
PYPROJECT = BASE / "pyproject.toml"
CHANGELOG = BASE / "docs" / "changelog.rst"

MD_CHANGELOG_SECTION_LIST = re.compile(r"- .+?(?=\n\n###|$)", re.DOTALL)
version_header = r"\d+\.\d+\.\d+ \([^)]+\)"
RST_LATEST_CHANGES = re.compile(
    rf"{version_header}\n--+\s+(.+?)\n\n+{version_header}", re.DOTALL
)


def update_docs_config(text: str, new: Version) -> str:
    new_major_minor = f"{new.major}.{new.minor}"
    text = re.sub(r"(?<=version = )[^\n]+", f'"{new_major_minor}"', text)
    return re.sub(r"(?<=release = )[^\n]+", f'"{new}"', text)


def update_changelog(text: str, new: Version) -> str:
    new_header = f"{new} ({datetime.now(timezone.utc).date():%B %d, %Y})"
    return re.sub(
        # do not match if the new version is already present
        r"\nUnreleased\n--+\n",
        rf"""
Unreleased
----------

New features:
Bug fixes:
For packagers:
Other changes:

{new_header}
{'-' * len(new_header)}
""",
        text,
    )


UpdateVersionCallable = Callable[[str, Version], str]
FILENAME_AND_UPDATE_TEXT: list[tuple[Path, UpdateVersionCallable]] = [
    (
        PYPROJECT,
        lambda text, new: re.sub(r"(?<=\nversion = )[^\n]+", f'"{new}"', text),
    ),
    (
        BASE / "beets" / "__init__.py",
        lambda text, new: re.sub(
            r"(?<=__version__ = )[^\n]+", f'"{new}"', text
        ),
    ),
    (CHANGELOG, update_changelog),
    (BASE / "docs" / "conf.py", update_docs_config),
]


def validate_new_version(
    ctx: click.Context, param: click.Argument, value: Version
) -> Version:
    """Validate the version is newer than the current one."""
    with PYPROJECT.open("rb") as f:
        current = parse(tomli.load(f)["tool"]["poetry"]["version"])

    if not value > current:
        msg = f"version must be newer than {current}"
        raise click.BadParameter(msg)

    return value


def bump_version(new: Version) -> None:
    """Update the version number in specified files."""
    for path, perform_update in FILENAME_AND_UPDATE_TEXT:
        with path.open("r+") as f:
            contents = f.read()
            f.seek(0)
            f.write(perform_update(contents, new))
            f.truncate()


def rst2md(text: str) -> str:
    """Use Pandoc to convert text from ReST to Markdown."""
    # Other backslashes with verbatim ranges.
    rst = re.sub(r"(?<=[\s(])`([^`]+)`(?=[^_])", r"``\1``", text)

    # Bug numbers.
    rst = re.sub(r":bug:`(\d+)`", r":bug: (#\1)", rst)

    # Users.
    rst = re.sub(r":user:`(\w+)`", r"@\1", rst)
    return (
        subprocess.check_output(
            ["/usr/bin/pandoc", "--from=rst", "--to=gfm", "--wrap=none"],
            input=rst.encode(),
        )
        .decode()
        .strip()
    )


def changelog_as_markdown() -> str:
    """Get the latest changelog entry as hacked up Markdown."""
    with CHANGELOG.open() as f:
        contents = f.read()

    m = RST_LATEST_CHANGES.search(contents)
    rst = m.group(1) if m else ""

    # Convert with Pandoc.
    md = rst2md(rst)

    # Make sections stand out
    md = re.sub(r"^(\w.+?):$", r"### \1", md, flags=re.M)

    # Highlight plugin names
    md = re.sub(
        r"^- `/?plugins/(\w+)`:?", r"- Plugin **`\1`**:", md, flags=re.M
    )

    # Highlights command names.
    md = re.sub(r"^- `(\w+)-cmd`:?", r"- Command **`\1`**:", md, flags=re.M)

    # sort list items alphabetically for each of the sections
    return MD_CHANGELOG_SECTION_LIST.sub(
        lambda m: "\n".join(sorted(m.group().splitlines())), md
    )


@click.group()
def cli():
    pass


@cli.command()
@click.argument("version", type=Version, callback=validate_new_version)
def bump(version: Version) -> None:
    """Bump the version in project files."""
    bump_version(version)


@cli.command()
def changelog():
    """Get the most recent version's changelog as Markdown."""
    print(changelog_as_markdown())


if __name__ == "__main__":
    cli()



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for autotagging functionality."""

import re
import unittest

import pytest

from beets import autotag, config
from beets.autotag import AlbumInfo, TrackInfo, match
from beets.autotag.hooks import Distance, string_dist
from beets.library import Item
from beets.test.helper import BeetsTestCase
from beets.util import plurality


class PluralityTest(BeetsTestCase):
    def test_plurality_consensus(self):
        objs = [1, 1, 1, 1]
        obj, freq = plurality(objs)
        assert obj == 1
        assert freq == 4

    def test_plurality_near_consensus(self):
        objs = [1, 1, 2, 1]
        obj, freq = plurality(objs)
        assert obj == 1
        assert freq == 3

    def test_plurality_conflict(self):
        objs = [1, 1, 2, 2, 3]
        obj, freq = plurality(objs)
        assert obj in (1, 2)
        assert freq == 2

    def test_plurality_empty_sequence_raises_error(self):
        with pytest.raises(ValueError, match="must be non-empty"):
            plurality([])

    def test_current_metadata_finds_pluralities(self):
        items = [
            Item(artist="The Beetles", album="The White Album"),
            Item(artist="The Beatles", album="The White Album"),
            Item(artist="The Beatles", album="Teh White Album"),
        ]
        likelies, consensus = match.current_metadata(items)
        assert likelies["artist"] == "The Beatles"
        assert likelies["album"] == "The White Album"
        assert not consensus["artist"]

    def test_current_metadata_artist_consensus(self):
        items = [
            Item(artist="The Beatles", album="The White Album"),
            Item(artist="The Beatles", album="The White Album"),
            Item(artist="The Beatles", album="Teh White Album"),
        ]
        likelies, consensus = match.current_metadata(items)
        assert likelies["artist"] == "The Beatles"
        assert likelies["album"] == "The White Album"
        assert consensus["artist"]

    def test_albumartist_consensus(self):
        items = [
            Item(artist="tartist1", album="album", albumartist="aartist"),
            Item(artist="tartist2", album="album", albumartist="aartist"),
            Item(artist="tartist3", album="album", albumartist="aartist"),
        ]
        likelies, consensus = match.current_metadata(items)
        assert likelies["artist"] == "aartist"
        assert not consensus["artist"]

    def test_current_metadata_likelies(self):
        fields = [
            "artist",
            "album",
            "albumartist",
            "year",
            "disctotal",
            "mb_albumid",
            "label",
            "barcode",
            "catalognum",
            "country",
            "media",
            "albumdisambig",
        ]
        items = [Item(**{f: f"{f}_{i or 1}" for f in fields}) for i in range(5)]
        likelies, _ = match.current_metadata(items)
        for f in fields:
            if isinstance(likelies[f], int):
                assert likelies[f] == 0
            else:
                assert likelies[f] == f"{f}_1"


def _make_item(title, track, artist="some artist"):
    return Item(
        title=title,
        track=track,
        artist=artist,
        album="some album",
        length=1,
        mb_trackid="",
        mb_albumid="",
        mb_artistid="",
    )


def _make_trackinfo():
    return [
        TrackInfo(
            title="one", track_id=None, artist="some artist", length=1, index=1
        ),
        TrackInfo(
            title="two", track_id=None, artist="some artist", length=1, index=2
        ),
        TrackInfo(
            title="three",
            track_id=None,
            artist="some artist",
            length=1,
            index=3,
        ),
    ]


def _clear_weights():
    """Hack around the lazy descriptor used to cache weights for
    Distance calculations.
    """
    Distance.__dict__["_weights"].cache = {}


class DistanceTest(BeetsTestCase):
    def tearDown(self):
        super().tearDown()
        _clear_weights()

    def test_add(self):
        dist = Distance()
        dist.add("add", 1.0)
        assert dist._penalties == {"add": [1.0]}

    def test_add_equality(self):
        dist = Distance()
        dist.add_equality("equality", "ghi", ["abc", "def", "ghi"])
        assert dist._penalties["equality"] == [0.0]

        dist.add_equality("equality", "xyz", ["abc", "def", "ghi"])
        assert dist._penalties["equality"] == [0.0, 1.0]

        dist.add_equality("equality", "abc", re.compile(r"ABC", re.I))
        assert dist._penalties["equality"] == [0.0, 1.0, 0.0]

    def test_add_expr(self):
        dist = Distance()
        dist.add_expr("expr", True)
        assert dist._penalties["expr"] == [1.0]

        dist.add_expr("expr", False)
        assert dist._penalties["expr"] == [1.0, 0.0]

    def test_add_number(self):
        dist = Distance()
        # Add a full penalty for each number of difference between two numbers.

        dist.add_number("number", 1, 1)
        assert dist._penalties["number"] == [0.0]

        dist.add_number("number", 1, 2)
        assert dist._penalties["number"] == [0.0, 1.0]

        dist.add_number("number", 2, 1)
        assert dist._penalties["number"] == [0.0, 1.0, 1.0]

        dist.add_number("number", -1, 2)
        assert dist._penalties["number"] == [0.0, 1.0, 1.0, 1.0, 1.0, 1.0]

    def test_add_priority(self):
        dist = Distance()
        dist.add_priority("priority", "abc", "abc")
        assert dist._penalties["priority"] == [0.0]

        dist.add_priority("priority", "def", ["abc", "def"])
        assert dist._penalties["priority"] == [0.0, 0.5]

        dist.add_priority(
            "priority", "gh", ["ab", "cd", "ef", re.compile("GH", re.I)]
        )
        assert dist._penalties["priority"] == [0.0, 0.5, 0.75]

        dist.add_priority("priority", "xyz", ["abc", "def"])
        assert dist._penalties["priority"] == [0.0, 0.5, 0.75, 1.0]

    def test_add_ratio(self):
        dist = Distance()
        dist.add_ratio("ratio", 25, 100)
        assert dist._penalties["ratio"] == [0.25]

        dist.add_ratio("ratio", 10, 5)
        assert dist._penalties["ratio"] == [0.25, 1.0]

        dist.add_ratio("ratio", -5, 5)
        assert dist._penalties["ratio"] == [0.25, 1.0, 0.0]

        dist.add_ratio("ratio", 5, 0)
        assert dist._penalties["ratio"] == [0.25, 1.0, 0.0, 0.0]

    def test_add_string(self):
        dist = Distance()
        sdist = string_dist("abc", "bcd")
        dist.add_string("string", "abc", "bcd")
        assert dist._penalties["string"] == [sdist]
        assert dist._penalties["string"] != [0]

    def test_add_string_none(self):
        dist = Distance()
        dist.add_string("string", None, "string")
        assert dist._penalties["string"] == [1]

    def test_add_string_both_none(self):
        dist = Distance()
        dist.add_string("string", None, None)
        assert dist._penalties["string"] == [0]

    def test_distance(self):
        config["match"]["distance_weights"]["album"] = 2.0
        config["match"]["distance_weights"]["medium"] = 1.0
        _clear_weights()

        dist = Distance()
        dist.add("album", 0.5)
        dist.add("media", 0.25)
        dist.add("media", 0.75)
        assert dist.distance == 0.5

        # __getitem__()
        assert dist["album"] == 0.25
        assert dist["media"] == 0.25

    def test_max_distance(self):
        config["match"]["distance_weights"]["album"] = 3.0
        config["match"]["distance_weights"]["medium"] = 1.0
        _clear_weights()

        dist = Distance()
        dist.add("album", 0.5)
        dist.add("medium", 0.0)
        dist.add("medium", 0.0)
        assert dist.max_distance == 5.0

    def test_operators(self):
        config["match"]["distance_weights"]["source"] = 1.0
        config["match"]["distance_weights"]["album"] = 2.0
        config["match"]["distance_weights"]["medium"] = 1.0
        _clear_weights()

        dist = Distance()
        dist.add("source", 0.0)
        dist.add("album", 0.5)
        dist.add("medium", 0.25)
        dist.add("medium", 0.75)
        assert len(dist) == 2
        assert list(dist) == [("album", 0.2), ("medium", 0.2)]
        assert dist == 0.4
        assert dist < 1.0
        assert dist > 0.0
        assert dist - 0.4 == 0.0
        assert 0.4 - dist == 0.0
        assert float(dist) == 0.4

    def test_raw_distance(self):
        config["match"]["distance_weights"]["album"] = 3.0
        config["match"]["distance_weights"]["medium"] = 1.0
        _clear_weights()

        dist = Distance()
        dist.add("album", 0.5)
        dist.add("medium", 0.25)
        dist.add("medium", 0.5)
        assert dist.raw_distance == 2.25

    def test_items(self):
        config["match"]["distance_weights"]["album"] = 4.0
        config["match"]["distance_weights"]["medium"] = 2.0
        _clear_weights()

        dist = Distance()
        dist.add("album", 0.1875)
        dist.add("medium", 0.75)
        assert dist.items() == [("medium", 0.25), ("album", 0.125)]

        # Sort by key if distance is equal.
        dist = Distance()
        dist.add("album", 0.375)
        dist.add("medium", 0.75)
        assert dist.items() == [("album", 0.25), ("medium", 0.25)]

    def test_update(self):
        dist1 = Distance()
        dist1.add("album", 0.5)
        dist1.add("media", 1.0)

        dist2 = Distance()
        dist2.add("album", 0.75)
        dist2.add("album", 0.25)
        dist2.add("media", 0.05)

        dist1.update(dist2)

        assert dist1._penalties == {
            "album": [0.5, 0.75, 0.25],
            "media": [1.0, 0.05],
        }


class TrackDistanceTest(BeetsTestCase):
    def test_identical_tracks(self):
        item = _make_item("one", 1)
        info = _make_trackinfo()[0]
        dist = match.track_distance(item, info, incl_artist=True)
        assert dist == 0.0

    def test_different_title(self):
        item = _make_item("foo", 1)
        info = _make_trackinfo()[0]
        dist = match.track_distance(item, info, incl_artist=True)
        assert dist != 0.0

    def test_different_artist(self):
        item = _make_item("one", 1)
        item.artist = "foo"
        info = _make_trackinfo()[0]
        dist = match.track_distance(item, info, incl_artist=True)
        assert dist != 0.0

    def test_various_artists_tolerated(self):
        item = _make_item("one", 1)
        item.artist = "Various Artists"
        info = _make_trackinfo()[0]
        dist = match.track_distance(item, info, incl_artist=True)
        assert dist == 0.0


class AlbumDistanceTest(BeetsTestCase):
    def _mapping(self, items, info):
        out = {}
        for i, t in zip(items, info.tracks):
            out[i] = t
        return out

    def _dist(self, items, info):
        return match.distance(items, info, self._mapping(items, info))

    def test_identical_albums(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        assert self._dist(items, info) == 0

    def test_incomplete_album(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        dist = self._dist(items, info)
        assert dist != 0
        # Make sure the distance is not too great
        assert dist < 0.2

    def test_global_artists_differ(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="someone else",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        assert self._dist(items, info) != 0

    def test_comp_track_artists_match(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="should be ignored",
            album="some album",
            tracks=_make_trackinfo(),
            va=True,
        )
        assert self._dist(items, info) == 0

    def test_comp_no_track_artists(self):
        # Some VA releases don't have track artists (incomplete metadata).
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="should be ignored",
            album="some album",
            tracks=_make_trackinfo(),
            va=True,
        )
        info.tracks[0].artist = None
        info.tracks[1].artist = None
        info.tracks[2].artist = None
        assert self._dist(items, info) == 0

    def test_comp_track_artists_do_not_match(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2, "someone else"))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=True,
        )
        assert self._dist(items, info) != 0

    def test_tracks_out_of_order(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("three", 2))
        items.append(_make_item("two", 3))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        dist = self._dist(items, info)
        assert 0 < dist < 0.2

    def test_two_medium_release(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 3))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        info.tracks[0].medium_index = 1
        info.tracks[1].medium_index = 2
        info.tracks[2].medium_index = 1
        dist = self._dist(items, info)
        assert dist == 0

    def test_per_medium_track_numbers(self):
        items = []
        items.append(_make_item("one", 1))
        items.append(_make_item("two", 2))
        items.append(_make_item("three", 1))
        info = AlbumInfo(
            artist="some artist",
            album="some album",
            tracks=_make_trackinfo(),
            va=False,
        )
        info.tracks[0].medium_index = 1
        info.tracks[1].medium_index = 2
        info.tracks[2].medium_index = 1
        dist = self._dist(items, info)
        assert dist == 0


class AssignmentTest(unittest.TestCase):
    def item(self, title, track):
        return Item(
            title=title,
            track=track,
            mb_trackid="",
            mb_albumid="",
            mb_artistid="",
        )

    def test_reorder_when_track_numbers_incorrect(self):
        items = []
        items.append(self.item("one", 1))
        items.append(self.item("three", 2))
        items.append(self.item("two", 3))
        trackinfo = []
        trackinfo.append(TrackInfo(title="one"))
        trackinfo.append(TrackInfo(title="two"))
        trackinfo.append(TrackInfo(title="three"))
        mapping, extra_items, extra_tracks = match.assign_items(
            items, trackinfo
        )
        assert extra_items == []
        assert extra_tracks == []
        assert mapping == {
            items[0]: trackinfo[0],
            items[1]: trackinfo[2],
            items[2]: trackinfo[1],
        }

    def test_order_works_with_invalid_track_numbers(self):
        items = []
        items.append(self.item("one", 1))
        items.append(self.item("three", 1))
        items.append(self.item("two", 1))
        trackinfo = []
        trackinfo.append(TrackInfo(title="one"))
        trackinfo.append(TrackInfo(title="two"))
        trackinfo.append(TrackInfo(title="three"))
        mapping, extra_items, extra_tracks = match.assign_items(
            items, trackinfo
        )
        assert extra_items == []
        assert extra_tracks == []
        assert mapping == {
            items[0]: trackinfo[0],
            items[1]: trackinfo[2],
            items[2]: trackinfo[1],
        }

    def test_order_works_with_missing_tracks(self):
        items = []
        items.append(self.item("one", 1))
        items.append(self.item("three", 3))
        trackinfo = []
        trackinfo.append(TrackInfo(title="one"))
        trackinfo.append(TrackInfo(title="two"))
        trackinfo.append(TrackInfo(title="three"))
        mapping, extra_items, extra_tracks = match.assign_items(
            items, trackinfo
        )
        assert extra_items == []
        assert extra_tracks == [trackinfo[1]]
        assert mapping == {items[0]: trackinfo[0], items[1]: trackinfo[2]}

    def test_order_works_with_extra_tracks(self):
        items = []
        items.append(self.item("one", 1))
        items.append(self.item("two", 2))
        items.append(self.item("three", 3))
        trackinfo = []
        trackinfo.append(TrackInfo(title="one"))
        trackinfo.append(TrackInfo(title="three"))
        mapping, extra_items, extra_tracks = match.assign_items(
            items, trackinfo
        )
        assert extra_items == [items[1]]
        assert extra_tracks == []
        assert mapping == {items[0]: trackinfo[0], items[2]: trackinfo[1]}

    def test_order_works_when_track_names_are_entirely_wrong(self):
        # A real-world test case contributed by a user.
        def item(i, length):
            return Item(
                artist="ben harper",
                album="burn to shine",
                title=f"ben harper - Burn to Shine {i}",
                track=i,
                length=length,
                mb_trackid="",
                mb_albumid="",
                mb_artistid="",
            )

        items = []
        items.append(item(1, 241.37243007106997))
        items.append(item(2, 342.27781704375036))
        items.append(item(3, 245.95070222338137))
        items.append(item(4, 472.87662515485437))
        items.append(item(5, 279.1759535763187))
        items.append(item(6, 270.33333768012))
        items.append(item(7, 247.83435613222923))
        items.append(item(8, 216.54504531525072))
        items.append(item(9, 225.72775379800484))
        items.append(item(10, 317.7643606963552))
        items.append(item(11, 243.57001238834192))
        items.append(item(12, 186.45916150485752))

        def info(index, title, length):
            return TrackInfo(title=title, length=length, index=index)

        trackinfo = []
        trackinfo.append(info(1, "Alone", 238.893))
        trackinfo.append(info(2, "The Woman in You", 341.44))
        trackinfo.append(info(3, "Less", 245.59999999999999))
        trackinfo.append(info(4, "Two Hands of a Prayer", 470.49299999999999))
        trackinfo.append(info(5, "Please Bleed", 277.86599999999999))
        trackinfo.append(info(6, "Suzie Blue", 269.30599999999998))
        trackinfo.append(info(7, "Steal My Kisses", 245.36000000000001))
        trackinfo.append(info(8, "Burn to Shine", 214.90600000000001))
        trackinfo.append(info(9, "Show Me a Little Shame", 224.0929999999999))
        trackinfo.append(info(10, "Forgiven", 317.19999999999999))
        trackinfo.append(info(11, "Beloved One", 243.733))
        trackinfo.append(info(12, "In the Lord's Arms", 186.13300000000001))

        mapping, extra_items, extra_tracks = match.assign_items(
            items, trackinfo
        )
        assert extra_items == []
        assert extra_tracks == []
        for item, info in mapping.items():
            assert items.index(item) == trackinfo.index(info)


class ApplyTestUtil:
    def _apply(self, info=None, per_disc_numbering=False, artist_credit=False):
        info = info or self.info
        mapping = {}
        for i, t in zip(self.items, info.tracks):
            mapping[i] = t
        config["per_disc_numbering"] = per_disc_numbering
        config["artist_credit"] = artist_credit
        autotag.apply_metadata(info, mapping)


class ApplyTest(BeetsTestCase, ApplyTestUtil):
    def setUp(self):
        super().setUp()

        self.items = []
        self.items.append(Item({}))
        self.items.append(Item({}))
        trackinfo = []
        trackinfo.append(
            TrackInfo(
                title="oneNew",
                track_id="dfa939ec-118c-4d0f-84a0-60f3d1e6522c",
                medium=1,
                medium_index=1,
                medium_total=1,
                index=1,
                artist_credit="trackArtistCredit",
                artists_credit=["trackArtistCredit"],
                artist_sort="trackArtistSort",
                artists_sort=["trackArtistSort"],
            )
        )
        trackinfo.append(
            TrackInfo(
                title="twoNew",
                track_id="40130ed1-a27c-42fd-a328-1ebefb6caef4",
                medium=2,
                medium_index=1,
                index=2,
                medium_total=1,
            )
        )
        self.info = AlbumInfo(
            tracks=trackinfo,
            artist="artistNew",
            artists=["artistNew", "artistNew2"],
            album="albumNew",
            album_id="7edb51cb-77d6-4416-a23c-3a8c2994a2c7",
            artist_id="a6623d39-2d8e-4f70-8242-0a9553b91e50",
            artists_ids=[
                "a6623d39-2d8e-4f70-8242-0a9553b91e50",
                "a6623d39-2d8e-4f70-8242-0a9553b91e51",
            ],
            artist_credit="albumArtistCredit",
            artists_credit=["albumArtistCredit", "albumArtistCredit2"],
            artist_sort="albumArtistSort",
            artists_sort=["albumArtistSort", "albumArtistSort2"],
            albumtype="album",
            va=False,
            mediums=2,
        )

    def test_titles_applied(self):
        self._apply()
        assert self.items[0].title == "oneNew"
        assert self.items[1].title == "twoNew"

    def test_album_and_artist_applied_to_all(self):
        self._apply()
        assert self.items[0].album == "albumNew"
        assert self.items[1].album == "albumNew"
        assert self.items[0].artist == "artistNew"
        assert self.items[1].artist == "artistNew"
        assert self.items[0].artists == ["artistNew", "artistNew2"]
        assert self.items[1].artists == ["artistNew", "artistNew2"]
        assert self.items[0].albumartists == ["artistNew", "artistNew2"]
        assert self.items[1].albumartists == ["artistNew", "artistNew2"]

    def test_track_index_applied(self):
        self._apply()
        assert self.items[0].track == 1
        assert self.items[1].track == 2

    def test_track_total_applied(self):
        self._apply()
        assert self.items[0].tracktotal == 2
        assert self.items[1].tracktotal == 2

    def test_disc_index_applied(self):
        self._apply()
        assert self.items[0].disc == 1
        assert self.items[1].disc == 2

    def test_disc_total_applied(self):
        self._apply()
        assert self.items[0].disctotal == 2
        assert self.items[1].disctotal == 2

    def test_per_disc_numbering(self):
        self._apply(per_disc_numbering=True)
        assert self.items[0].track == 1
        assert self.items[1].track == 1

    def test_per_disc_numbering_track_total(self):
        self._apply(per_disc_numbering=True)
        assert self.items[0].tracktotal == 1
        assert self.items[1].tracktotal == 1

    def test_artist_credit(self):
        self._apply(artist_credit=True)
        assert self.items[0].artist == "trackArtistCredit"
        assert self.items[1].artist == "albumArtistCredit"
        assert self.items[0].albumartist == "albumArtistCredit"
        assert self.items[1].albumartist == "albumArtistCredit"
        assert self.items[0].albumartists == [
            "albumArtistCredit",
            "albumArtistCredit2",
        ]
        assert self.items[1].albumartists == [
            "albumArtistCredit",
            "albumArtistCredit2",
        ]

    def test_artist_credit_prefers_artist_over_albumartist_credit(self):
        self.info.tracks[0].artist = "oldArtist"
        self.info.tracks[0].artist_credit = None
        self._apply(artist_credit=True)
        assert self.items[0].artist == "oldArtist"

    def test_artist_credit_falls_back_to_albumartist(self):
        self.info.artist_credit = None
        self._apply(artist_credit=True)
        assert self.items[1].artist == "artistNew"

    def test_mb_trackid_applied(self):
        self._apply()
        assert (
            self.items[0].mb_trackid == "dfa939ec-118c-4d0f-84a0-60f3d1e6522c"
        )
        assert (
            self.items[1].mb_trackid == "40130ed1-a27c-42fd-a328-1ebefb6caef4"
        )

    def test_mb_albumid_and_artistid_applied(self):
        self._apply()
        for item in self.items:
            assert item.mb_albumid == "7edb51cb-77d6-4416-a23c-3a8c2994a2c7"
            assert item.mb_artistid == "a6623d39-2d8e-4f70-8242-0a9553b91e50"
            assert item.mb_artistids == [
                "a6623d39-2d8e-4f70-8242-0a9553b91e50",
                "a6623d39-2d8e-4f70-8242-0a9553b91e51",
            ]

    def test_albumtype_applied(self):
        self._apply()
        assert self.items[0].albumtype == "album"
        assert self.items[1].albumtype == "album"

    def test_album_artist_overrides_empty_track_artist(self):
        my_info = self.info.copy()
        self._apply(info=my_info)
        assert self.items[0].artist == "artistNew"
        assert self.items[1].artist == "artistNew"
        assert self.items[0].artists == ["artistNew", "artistNew2"]
        assert self.items[1].artists == ["artistNew", "artistNew2"]

    def test_album_artist_overridden_by_nonempty_track_artist(self):
        my_info = self.info.copy()
        my_info.tracks[0].artist = "artist1!"
        my_info.tracks[1].artist = "artist2!"
        my_info.tracks[0].artists = ["artist1!", "artist1!!"]
        my_info.tracks[1].artists = ["artist2!", "artist2!!"]
        self._apply(info=my_info)
        assert self.items[0].artist == "artist1!"
        assert self.items[1].artist == "artist2!"
        assert self.items[0].artists == ["artist1!", "artist1!!"]
        assert self.items[1].artists == ["artist2!", "artist2!!"]

    def test_artist_credit_applied(self):
        self._apply()
        assert self.items[0].albumartist_credit == "albumArtistCredit"
        assert self.items[0].albumartists_credit == [
            "albumArtistCredit",
            "albumArtistCredit2",
        ]
        assert self.items[0].artist_credit == "trackArtistCredit"
        assert self.items[0].artists_credit == ["trackArtistCredit"]
        assert self.items[1].albumartist_credit == "albumArtistCredit"
        assert self.items[1].albumartists_credit == [
            "albumArtistCredit",
            "albumArtistCredit2",
        ]
        assert self.items[1].artist_credit == "albumArtistCredit"
        assert self.items[1].artists_credit == [
            "albumArtistCredit",
            "albumArtistCredit2",
        ]

    def test_artist_sort_applied(self):
        self._apply()
        assert self.items[0].albumartist_sort == "albumArtistSort"
        assert self.items[0].albumartists_sort == [
            "albumArtistSort",
            "albumArtistSort2",
        ]
        assert self.items[0].artist_sort == "trackArtistSort"
        assert self.items[0].artists_sort == ["trackArtistSort"]
        assert self.items[1].albumartist_sort == "albumArtistSort"
        assert self.items[1].albumartists_sort == [
            "albumArtistSort",
            "albumArtistSort2",
        ]
        assert self.items[1].artist_sort == "albumArtistSort"
        assert self.items[1].artists_sort == [
            "albumArtistSort",
            "albumArtistSort2",
        ]

    def test_full_date_applied(self):
        my_info = self.info.copy()
        my_info.year = 2013
        my_info.month = 12
        my_info.day = 18
        self._apply(info=my_info)

        assert self.items[0].year == 2013
        assert self.items[0].month == 12
        assert self.items[0].day == 18

    def test_date_only_zeros_month_and_day(self):
        self.items = []
        self.items.append(Item(year=1, month=2, day=3))
        self.items.append(Item(year=4, month=5, day=6))

        my_info = self.info.copy()
        my_info.year = 2013
        self._apply(info=my_info)

        assert self.items[0].year == 2013
        assert self.items[0].month == 0
        assert self.items[0].day == 0

    def test_missing_date_applies_nothing(self):
        self.items = []
        self.items.append(Item(year=1, month=2, day=3))
        self.items.append(Item(year=4, month=5, day=6))

        self._apply()

        assert self.items[0].year == 1
        assert self.items[0].month == 2
        assert self.items[0].day == 3

    def test_data_source_applied(self):
        my_info = self.info.copy()
        my_info.data_source = "MusicBrainz"
        self._apply(info=my_info)

        assert self.items[0].data_source == "MusicBrainz"


class ApplyCompilationTest(BeetsTestCase, ApplyTestUtil):
    def setUp(self):
        super().setUp()

        self.items = []
        self.items.append(Item({}))
        self.items.append(Item({}))
        trackinfo = []
        trackinfo.append(
            TrackInfo(
                title="oneNew",
                track_id="dfa939ec-118c-4d0f-84a0-60f3d1e6522c",
                artist="artistOneNew",
                artist_id="a05686fc-9db2-4c23-b99e-77f5db3e5282",
                index=1,
            )
        )
        trackinfo.append(
            TrackInfo(
                title="twoNew",
                track_id="40130ed1-a27c-42fd-a328-1ebefb6caef4",
                artist="artistTwoNew",
                artist_id="80b3cf5e-18fe-4c59-98c7-e5bb87210710",
                index=2,
            )
        )
        self.info = AlbumInfo(
            tracks=trackinfo,
            artist="variousNew",
            album="albumNew",
            album_id="3b69ea40-39b8-487f-8818-04b6eff8c21a",
            artist_id="89ad4ac3-39f7-470e-963a-56509c546377",
            albumtype="compilation",
        )

    def test_album_and_track_artists_separate(self):
        self._apply()
        assert self.items[0].artist == "artistOneNew"
        assert self.items[1].artist == "artistTwoNew"
        assert self.items[0].albumartist == "variousNew"
        assert self.items[1].albumartist == "variousNew"

    def test_mb_albumartistid_applied(self):
        self._apply()
        assert (
            self.items[0].mb_albumartistid
            == "89ad4ac3-39f7-470e-963a-56509c546377"
        )
        assert (
            self.items[1].mb_albumartistid
            == "89ad4ac3-39f7-470e-963a-56509c546377"
        )
        assert (
            self.items[0].mb_artistid == "a05686fc-9db2-4c23-b99e-77f5db3e5282"
        )
        assert (
            self.items[1].mb_artistid == "80b3cf5e-18fe-4c59-98c7-e5bb87210710"
        )

    def test_va_flag_cleared_does_not_set_comp(self):
        self._apply()
        assert not self.items[0].comp
        assert not self.items[1].comp

    def test_va_flag_sets_comp(self):
        va_info = self.info.copy()
        va_info.va = True
        self._apply(info=va_info)
        assert self.items[0].comp
        assert self.items[1].comp


class StringDistanceTest(unittest.TestCase):
    def test_equal_strings(self):
        dist = string_dist("Some String", "Some String")
        assert dist == 0.0

    def test_different_strings(self):
        dist = string_dist("Some String", "Totally Different")
        assert dist != 0.0

    def test_punctuation_ignored(self):
        dist = string_dist("Some String", "Some.String!")
        assert dist == 0.0

    def test_case_ignored(self):
        dist = string_dist("Some String", "sOME sTring")
        assert dist == 0.0

    def test_leading_the_has_lower_weight(self):
        dist1 = string_dist("XXX Band Name", "Band Name")
        dist2 = string_dist("The Band Name", "Band Name")
        assert dist2 < dist1

    def test_parens_have_lower_weight(self):
        dist1 = string_dist("One .Two.", "One")
        dist2 = string_dist("One (Two)", "One")
        assert dist2 < dist1

    def test_brackets_have_lower_weight(self):
        dist1 = string_dist("One .Two.", "One")
        dist2 = string_dist("One [Two]", "One")
        assert dist2 < dist1

    def test_ep_label_has_zero_weight(self):
        dist = string_dist("My Song (EP)", "My Song")
        assert dist == 0.0

    def test_featured_has_lower_weight(self):
        dist1 = string_dist("My Song blah Someone", "My Song")
        dist2 = string_dist("My Song feat Someone", "My Song")
        assert dist2 < dist1

    def test_postfix_the(self):
        dist = string_dist("The Song Title", "Song Title, The")
        assert dist == 0.0

    def test_postfix_a(self):
        dist = string_dist("A Song Title", "Song Title, A")
        assert dist == 0.0

    def test_postfix_an(self):
        dist = string_dist("An Album Title", "Album Title, An")
        assert dist == 0.0

    def test_empty_strings(self):
        dist = string_dist("", "")
        assert dist == 0.0

    def test_solo_pattern(self):
        # Just make sure these don't crash.
        string_dist("The ", "")
        string_dist("(EP)", "(EP)")
        string_dist(", An", "")

    def test_heuristic_does_not_harm_distance(self):
        dist = string_dist("Untitled", "[Untitled]")
        assert dist == 0.0

    def test_ampersand_expansion(self):
        dist = string_dist("And", "&")
        assert dist == 0.0

    def test_accented_characters(self):
        dist = string_dist("\xe9\xe1\xf1", "ean")
        assert dist == 0.0



"""Tests for the 'permissions' plugin."""

import os
import platform
from unittest.mock import Mock, patch

from beets.test._common import touch
from beets.test.helper import AsIsImporterMixin, ImportTestCase, PluginMixin
from beets.util import displayable_path
from beetsplug.permissions import (
    check_permissions,
    convert_perm,
    dirs_in_library,
)


class PermissionsPluginTest(AsIsImporterMixin, PluginMixin, ImportTestCase):
    plugin = "permissions"

    def setUp(self):
        super().setUp()

        self.config["permissions"] = {"file": "777", "dir": "777"}

    def test_permissions_on_album_imported(self):
        self.do_thing(True)

    def test_permissions_on_item_imported(self):
        self.config["import"]["singletons"] = True
        self.do_thing(True)

    @patch("os.chmod", Mock())
    def test_failing_to_set_permissions(self):
        self.do_thing(False)

    def do_thing(self, expect_success):
        if platform.system() == "Windows":
            self.skipTest("permissions not available on Windows")

        def get_stat(v):
            return (
                os.stat(os.path.join(self.temp_dir, b"import", *v)).st_mode
                & 0o777
            )

        typs = ["file", "dir"]

        track_file = (b"album", b"track_1.mp3")
        self.exp_perms = {
            True: {
                k: convert_perm(self.config["permissions"][k].get())
                for k in typs
            },
            False: {k: get_stat(v) for (k, v) in zip(typs, (track_file, ()))},
        }

        self.run_asis_importer()
        item = self.lib.items().get()

        self.assertPerms(item.path, "file", expect_success)

        for path in dirs_in_library(self.lib.directory, item.path):
            self.assertPerms(path, "dir", expect_success)

    def assertPerms(self, path, typ, expect_success):
        for x in [
            (True, self.exp_perms[expect_success][typ], "!="),
            (False, self.exp_perms[not expect_success][typ], "=="),
        ]:
            msg = "{} : {} {} {}".format(
                displayable_path(path),
                oct(os.stat(path).st_mode),
                x[2],
                oct(x[1]),
            )
            assert x[0] == check_permissions(path, x[1]), msg

    def test_convert_perm_from_string(self):
        assert convert_perm("10") == 8

    def test_convert_perm_from_int(self):
        assert convert_perm(10) == 8

    def test_permissions_on_set_art(self):
        self.do_set_art(True)

    @patch("os.chmod", Mock())
    def test_failing_permissions_on_set_art(self):
        self.do_set_art(False)

    def do_set_art(self, expect_success):
        if platform.system() == "Windows":
            self.skipTest("permissions not available on Windows")
        self.run_asis_importer()
        album = self.lib.albums().get()
        artpath = os.path.join(self.temp_dir, b"cover.jpg")
        touch(artpath)
        album.set_art(artpath)
        assert expect_success == check_permissions(album.artpath, 0o777)
